# -*- coding: utf-8 -*-

from __future__ import unicode_literals

"""Lore doesn't have much of a UI. Text output should be excellent.

"""
RESET = 0

BOLD = 1
FAINT = 2
ITALIC = 3
UNDERLINE = 4
STROBE = 5
BLINK = 6
INVERSE = 7
CONCEAL = 8
STRIKE = 9

BLACK = 30
RED = 31
GREEN = 32
YELLOW = 33
BLUE = 34
MAGENTA = 35
CYAN = 36
WHITE = 37
DEFAULT = 39


def debug(content='DEBUG'):
    return gray(7, content)


def info(content='INFO'):
    return foreground(BLUE, content)


def warning(content='WARNING'):
    """ warning style
    
    :param content:
    :return: string
    """
    return foreground(MAGENTA, content)


def success(content='SUCCESS'):
    """ success style
    
    :param content:
    :return: string
    """
    return foreground(GREEN, content)


def error(content='ERROR'):
    """ error style
    
    :param content:
    :return: string
    """
    return foreground(RED, content)


def critical(content='CRITICAL'):
    return blink(foreground(bright(RED), content))


def foreground(color, content, readline=False):
    """ Color the text of the content
    
    :param color:
    :param content:
    :return:
    """
    return encode(color, readline=readline) + content + encode(DEFAULT, readline=readline)


def background(color, content):
    """ Color the background of the content
    
    :param color:
    :param content:
    :return:
    """
    return encode(color + 10) + content + encode(DEFAULT + 10)


def bright(color):
    """ Brighten colors
    
    :param color:
    :return: brighter version of the color
    """
    return color + 60


def gray(gray, content):
    """ Grayscale
    
    :param gray: [0-15] 0 is almost black, 15 is nearly white
    :param content: content
    :return: string
    """
    return encode('38;5;%i' % (232 + gray)) + content + encode(DEFAULT)


def rgb(red, green, blue, content):
    """ Colors a content using rgb for h
    :param red: [0-5]
    :param green: [0-5]
    :param blue: [0-5]
    :param content: content to markup
    :return: string
    """
    rgb = 16 + 36 * red + 6 * green + blue
    return encode('38;5;' + str(rgb)) + content + encode(DEFAULT)


def bold(content):
    """ Bold content
    
    :param content:
    :return: string
    """
    return style(BOLD, content)


def faint(content):
    """ Faint content
    
    :param content:
    :return: string
    """
    return style(FAINT, content)


def italic(content):
    """ Italic content
    
    :param content:
    :return: string
    """
    return style(ITALIC, content)


def underline(content):
    """ Underline content
    
    :param content:
    :return: string
    """
    return style(UNDERLINE, content)


def strobe(content):
    """ Quickly blinking content
    
    :param content:
    :return: string
    """
    return style(STROBE, content)


def blink(content):
    """ Slowing blinking content
    
    :param content:
    :return: string
    """
    return style(BLINK, content)


def strike(content):
    """ Strike through content
    
    :param content:
    :return: string
    """
    return style(STRIKE, content)


def reset():
    """ Remove any active ansi styles
    
    :return: string resetting marker
    """
    return encode(RESET)


def style(effect, content):
    """ add a particular style to the content
    
    :param effect: style
    :param content: string
    :return: string
    """
    return encode(effect) + content + encode(effect + 20)


def encode(code, readline=False):
    """ Adds escape and control characters for ANSI codes
    
    :param code:
    :return:
    """
    if readline:
        return '\001\033[' + str(code) + 'm\002'

    return '\033[' + str(code) + 'm'



# -*- coding: utf-8 -*-

from lore.features.base import Base
from abc import ABCMeta, abstractmethod
from lore.io import upload
import json
import tempfile

class S3(Base):
    __metaclass__ = ABCMeta

    @abstractmethod
    def serialization(self):
        pass

    def publish(self):
        temp_file, temp_path = tempfile.mkstemp()
        data = self.get_data()

        if self.serialization() == 'csv':
            data.to_csv(temp_path, index=False)
        elif self.serialization() == 'pickle':
            data.to_pickle(temp_path)
        else:
            raise "Invalid serialization"
        upload(temp_path, self.data_path())

        with open(temp_path, 'w') as f:
            f.write(json.dumps(self.metadata()))
        upload(temp_path, self.metadata_path())

    def data_path(self):
        return "{}/{}/data.{}".format(self.version, self.name(), self.serialization())

    def metadata_path(self):
        return "{}/{}/metadata.json".format(self.version, self.name())



import pickle
import os
from os.path import join
import h5py
import json
import re
import warnings
import inspect

import lore
from lore import io
from lore.util import timer


class Base(object):
    def __init__(self, klass=None, model=None):
        frame, filename, line_number, function_name, lines, index = inspect.stack()[1]
        warnings.showwarning('Serializers are now directly implemented in models. Please use the model class directly instead"', DeprecationWarning,
                             filename, line_number)
        self.local_path = None
        self.remote_path = None
        self.fitting_path = None
        self.model_path = None
        self.remote_model_path = None
        self.model = model
        self._fitting = None

        if model is not None:
            self.local_path = join(lore.env.models_dir, model.__module__, model.__class__.__name__)
            self.remote_path = join(model.__module__, model.__class__.__name__)
        elif klass is not None:
            self.local_path = join(lore.env.models_dir, klass.__module__, klass.__name__)
            self.remote_path = join(klass.__module__, klass.__name__)
        else:
            raise ValueError('You must pass name or model')
        self.fitting = self.last_fitting()

    def last_fitting(self):
        if not os.path.exists(self.local_path):
            return 1
            
        fittings = [int(d) for d in os.listdir(self.local_path) if re.match(r'^\d+$', d)]
        if not fittings:
            return 1
        
        return sorted(fittings)[-1]

    @property
    def fitting(self):
        return self._fitting
    
    @fitting.setter
    def fitting(self, value):
        self._fitting = value
        self.fitting_path = join(self.local_path, str(self._fitting))
        model_file = 'model.pickle'
        self.model_path = join(self.fitting_path, model_file)
        self.remote_model_path = join(self.remote_path, model_file)
        if not os.path.exists(self.fitting_path):
            os.makedirs(self.fitting_path)

    def save(self, stats=None):
        with timer('pickle model:'):
            with open(self.model_path, 'wb') as f:
                pickle.dump(self.model, f)

        with open(join(self.fitting_path, 'params.json'), 'w') as f:
            if hasattr(self.model.estimator, 'params'):
                params = self.model.estimator.params
            else:
                params = {}
                for key, value in self.model.estimator.__getstate__().items():
                    if not key.startswith('_'):
                        params[key] = value.__repr__()
            json.dump(params, f, indent=2, sort_keys=True)

        if stats:
            with open(join(self.fitting_path, 'stats.json'), 'w') as f:
                json.dump(stats, f, indent=2, sort_keys=True)
        
    def load(self, fitting=None):
        if fitting:
            self.fitting = fitting
        
        with timer('unpickle model:'):
            with open(self.model_path, 'rb') as f:
                self.model = pickle.load(f)

    def upload(self):
        self.save()
        io.upload(self.model_path, self.remote_model_path)

    def download(self):
        self.fitting = 0
        io.download(self.model_path, self.remote_model_path)
        return self.load()


class Keras(Base):
    def __init__(self, klass=None, model=None):
        self.weights_path = None
        self.remote_weights_path = None
        self.checkpoint_path = None
        self.tensorboard_path = None
        self.timeline_path = None
        super(Keras, self).__init__(klass=klass, model=model)

    @Base.fitting.setter
    def fitting(self, value):
        Base.fitting.fset(self, value)  # this is a "pythonic" super() call :(
        weights_file = 'weights.h5'
        self.weights_path = join(self.fitting_path, weights_file)
        self.remote_weights_path = join(self.remote_path, weights_file)
        self.checkpoint_path = join(self.fitting_path, 'checkpoints/{epoch}.h5')
        if not os.path.exists(os.path.dirname(self.checkpoint_path)):
            os.makedirs(os.path.dirname(self.checkpoint_path))
        self.tensorboard_path = join(self.fitting_path, 'tensorboard')
        if not os.path.exists(os.path.dirname(self.tensorboard_path)):
            os.makedirs(os.path.dirname(self.tensorboard_path))
        self.timeline_path = join(self.fitting_path, 'timeline.json')
        if not os.path.exists(os.path.dirname(self.timeline_path)):
            os.makedirs(os.path.dirname(self.timeline_path))

    def save(self, stats=None):
        super(Keras, self).save(stats)
        
        with timer('save weights:'):
            # Only save weights, because saving named layers that have shared
            # weights causes an error on reload
            self.model.estimator.keras.save_weights(self.weights_path)

        # Patch for keras 2 models saved with optimizer weights:
        # https://github.com/gagnonlg/explore-ml/commit/c05b01076c7eb99dae6a480a05ac14765ef08e4b
        with h5py.File(self.weights_path, 'a') as f:
            if 'optimizer_weights' in f.keys():
                del f['optimizer_weights']
        
    def load(self, fitting=None):
        super(Keras, self).load(fitting)

        if hasattr(self.model, 'estimator'):
            # HACK to set estimator model, and model serializer
            self.model.estimator = self.model.estimator
    
            # Rely on build + load_weights rather than loading the named layers
            # w/ Keras for efficiency (and also because it causes a
            # deserialization issue) as of Keras 2.0.4:
            # https://github.com/fchollet/keras/issues/5442
            self.model.estimator.build()
            
            with timer('load weights:'):
                self.model.estimator.keras.load_weights(self.weights_path)
        else:
            self.model.build()
            with timer('load weights:'):
                self.model.keras.load_weights(self.weights_path)

        return self.model

    def upload(self):
        super(Keras, self).upload()
        io.upload(self.weights_path, self.remote_weights_path)

    def download(self):
        self.fitting = 0
        io.download(self.weights_path, self.remote_weights_path)
        return super(Keras, self).download()



import pandas
import hashlib
import logging
import os

from abc import abstractmethod, abstractproperty

import lore.env
from lore.util import timer, timed

logger = logging.getLogger(__name__)


class Table(object):

    def __init__(self, connection, sql, params, chunksize):
        self.connection = connection
        self.sql = sql
        self.params = params
        self.chunksize = chunksize
        self.key = 'sql_' + self.orientation + hashlib.sha1(sql.encode('utf-8') + str(params).encode('utf-8')).hexdigest()
        self.key_columns = self.key + '_columns'
        self.key_min_itemsize = self.key + '_min_itemsize'
        self.loaded = False
        self._store = None
        self._min_itemsize = None
        self._columns = None

    @abstractmethod
    def __len__(self):
        pass

    @property
    def columns(self):
        if not self.loaded:
            self.load()
    
        if self._columns is None:
            with self.store as store:
                self._columns = list(store[self.key_columns])
    
        return self._columns

    @property
    def min_itemsize(self):
        if not self.loaded:
            self.load()
    
        if self._min_itemsize is None:
            with self.store as store:
                self._min_itemsize = store[self.key_min_itemsize].to_dict()
    
        return self._min_itemsize

    @property
    def store(self):
        if self._store is None:
            self._store = pandas.HDFStore(os.path.join(lore.env.data_dir, self.connection.name + '.h5'), complevel=7,
                                          complib='blosc')
    
        if not self._store.is_open:
            self._store.open()
    
        return self._store

    def select(self, columns=None):
        if not self._store.is_open:
            self._store.open()
        return self._store.select(self.key, columns=columns, chunksize=self.chunksize)
        
    @abstractproperty
    def orientation(self):
        return None
    
    @abstractmethod
    def load(self):
        pass
    
    @abstractmethod
    def index(self, columns, optlevel=7, kind='full'):
        pass


class Rows(Table):
    """ Row oriented table in chunks """
    def __getitem__(self, name):
        with self.store as store:
            return store.select(self.key, columns=name, chunksize=self.chunksize)

    def __len__(self):
        with self.store as store:
            return store.get_storer(self.key).nrows

    @property
    def orientation(self):
        return 'row'
    
    @timed(logging.INFO)
    def load(self):
        self.loaded = True
        
        with self.store as store:
            if self.key in store:
                return
        
        self._min_itemsize = {}
        for dataframe in pandas.read_sql(self.sql, self.connection._connection, chunksize=self.chunksize,
                                         params=self.params):
            logger.info('appending %i rows to hdf5' % len(dataframe))
            self._columns = dataframe.columns
            
            with self.store as store:
                resize = False
                for column in dataframe.columns:
                    series = dataframe[column]
                    if series.dtype == 'object':
                        # resize string columns on disk if required
                        max = int(series.str.encode('utf-8').str.len().max())
                        if column not in self._min_itemsize:
                            self.min_itemsize[column] = max
                        
                        if max > self._min_itemsize[column]:
                            self.min_itemsize[column] = max
                            resize = True
                
                if resize:
                    with timer('resize %s' % str(self.min_itemsize), logging.INFO):
                        temp = 'resize_' + self.key
                        store.get_node(self.key)._f_rename(temp)
                        
                        for chunk in store.select(temp, chunksize=self.chunksize):
                            store.append(self.key, chunk, index=False, min_itemsize=self.min_itemsize)
                        del store[temp]
                
                store.append(self.key, dataframe, index=False)
        
        with self.store as store:
            store[self.key_columns] = pandas.Series(self.columns)
            store[self.key_min_itemsize] = pandas.Series(self.min_itemsize)
    
    def index(self, columns, optlevel=7, kind='full'):
        self.store.create_table_index(self.key, columns=columns, optlevel=optlevel, kind=kind)


class Columns(Table):
    """ Column oriented table in chunks """

    def __getitem__(self, name):
        with self.store as store:
            return store.select(self.column_key(name), chunksize=self.chunksize)

    def __len__(self):
        with self.store as store:
            return store.get_storer(self.column_key(self.columns[0])).nrows

    @property
    def orientation(self):
        return 'column'

    @timed(logging.INFO)
    def load(self):
        self.loaded = True

        with self.store as store:
            if self.key in store:
                return
    
        self._min_itemsize = {}
        for dataframe in pandas.read_sql(self.sql, self.connection._connection, chunksize=self.chunksize, params=self.params):
            logger.info('appending %i rows to hdf5' % len(dataframe))
            self._columns = dataframe.columns

            with self.store as store:
                for column in dataframe.columns:
                    name = self.column_key(column)
                    series = dataframe[column]

                    if name.lower() == 'index':
                        raise AttributeError("'index' is a reserved word and cannot be used for a column name")

                    if series.dtype == 'object':
                        # resize string columns on disk if required
                        max = int(series.str.encode('utf-8').str.len().max())
                        if column not in self._min_itemsize:
                            self._min_itemsize[column] = max
                
                        if max > self._min_itemsize[column]:
                            self._min_itemsize[column] = max
                            min_itemsize = {column: max}

                            with timer('resize %s' % str(min_itemsize), logging.INFO):
                                temp = 'resize_' + name
                                store.get_node(name)._f_rename(temp)
                                
                                for chunk in store.select(temp, chunksize=self.chunksize):
                                    store.append(name, chunk, index=False, min_itemsize=min_itemsize)
                                del store[temp]

                    store.append(name, series, index=False)

        with self.store as store:
            store[self.key_columns] = pandas.Series(self.columns)
            store[self.key_min_itemsize] = pandas.Series(self.min_itemsize)

    def column_key(self, column):
        return self.key + '_' + str(column)
    
    def index(self, columns, optlevel=7, kind='full'):
        for column in columns:
            self.store.create_table_index(self.column_key(column), columns=column, optlevel=optlevel, kind=kind)





import logging

from sklearn.base import BaseEstimator
import holtwinters
from lore.util import timed


class HoltWinters(BaseEstimator):

  def __init__(self, **kwargs):
    super(HoltWinters, self).__init__()
    self.periodicity = kwargs.get('periodicity')
    self.forecasts = kwargs.get('days_to_forecast')
    self.kwargs = kwargs
    self.params = None

  @timed(logging.INFO)
  def fit(self, x, y=None):
    results = holtwinters.additive(x, self.periodicity, self.forecasts, 
      alpha=self.kwargs.get('alpha'),
      beta=self.kwargs.get('beta'),
      gamma=self.kwargs.get('gamma'))
    self.params = {'alpha': results[1], 'beta': results[2], 'gamma': results[3]}
    self.rmse = results[4]
    return {'alpha': results[1], 'beta': results[2], 'gamma': results[3], 'RMSE': self.rmse}

  @timed(logging.INFO)
  def predict(self, X):
    return holtwinters.additive(X, self.periodicity, self.forecasts, 
      **self.params)[0]
    


from __future__ import unicode_literals
from abc import ABCMeta, abstractmethod

import csv
import datetime
import os
import re

import inflection
import numpy


class Base(object):
    __metaclass__ = ABCMeta
    
    def __init__(self, column):
        self.column = column
        self.name = self.column + '_' + inflection.underscore(self.__class__.__name__)
    
    @abstractmethod
    def transform(self, data):
        pass


class Map(Base):
    def transform(self, data):
        return data.loc[:, self.column].map(self.__class__.MAP)


class DateTime(Base):
    """
    For available operators see:
    https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DatetimeIndex.html#pandas.DatetimeIndex
    """
    
    def __init__(self, column, operator):
        super(DateTime, self).__init__(column)
        self.operator = operator
        self.name = self.column + '_' + inflection.underscore(self.__class__.__name__) + '_' + self.operator

    def transform(self, data):
        return getattr(data[self.column].dt, self.operator)


class Age(Base):
    def __init__(self, column, unit='seconds'):
        super(Age, self).__init__(column)
        self.unit = unit

    def transform(self, data):
        age = (datetime.datetime.now() - data[self.column])
        if self.unit in ['nanosecond', 'nanoseconds']:
            return age
        
        seconds = age.dt.total_seconds()
        if self.unit in ['second', 'seconds']:
            return seconds
        if self.unit in ['minute', 'minutes']:
            return seconds / 60
        if self.unit in ['hour', 'hours']:
            return seconds / 3600
        if self.unit in ['day', 'days']:
            return seconds / 86400
        if self.unit in ['week', 'weeks']:
            return seconds / 604800
        if self.unit in ['month', 'months']:
            return seconds / 2592000
        if self.unit in ['year', 'years']:
            return seconds / 31536000

        raise NameError('Unknown unit: %s' % self.unit)
        

class Log(Base):
    def transform(self, data):
        return numpy.log(data.loc[:, self.column])


class LogPlusOne(Base):
    def transform(self, data):
        return numpy.log1p(numpy.maximum(data.loc[:, self.column], 0))


class AreaCode(Base):
    """Transforms various phone number formats into area codes (strings)
    
    e.g. '12345678901' => '234'
         '+1 (234) 567-8901' => '234'
         '1234567' => ''
         float.nan => None
    """

    COUNTRY_DIGITS = re.compile(r'^\+?1(\d{10})$', re.UNICODE)
    PUNCTUATED = re.compile(r'(?:1[.\-]?)?\s?\(?(\d{3})\)?\s?[.\-]?[\d]{3}[.\-]?[\d]{4}', re.UNICODE)

    def transform(self, data):
        series = data[self.column].astype(object)
        countries = series.str.extract(AreaCode.COUNTRY_DIGITS, expand=False)
        countries = countries.str[0:3]
        punctuated = series.str.extract(AreaCode.PUNCTUATED, expand=False)
        areacodes = countries
        areacodes[areacodes.isnull()] = punctuated
        areacodes[areacodes.isnull()] = ''
        areacodes[series.isnull()] = None
        return areacodes


class EmailDomain(Base):
    """Transforms email addresses into their full domain name
    
    e.g. 'bob@bob.com' => 'bob.com'
    """
    NAIVE = re.compile(r'^[^@]+@(.+)$', re.UNICODE)

    def transform(self, data):
        domains = data[self.column].str.extract(EmailDomain.NAIVE, expand=False)
        domains[domains.isnull()] = ''
        return domains


class NameAge(Map):
    MAP = {}
    
    with open(os.path.join(os.path.dirname(__file__), 'data', 'names.csv'), 'r') as file:
        reader = csv.reader(file)
        for line in reader:
            MAP[line[0]] = float(line[2])


class NamePopulation(Map):
    MAP = {}
    
    with open(os.path.join(os.path.dirname(__file__), 'data', 'names.csv'), 'r') as file:
        reader = csv.reader(file)
        for line in reader:
            MAP[line[0]] = float(line[3])


class NameSex(Map):
    MAP = {}
    
    with open(os.path.join(os.path.dirname(__file__), 'data', 'names.csv'), 'r') as file:
        reader = csv.reader(file)
        for line in reader:
            MAP[line[0]] = float(line[1])


class NameFamilial(Base):
    NAIVE = re.compile(r'\b(mom|dad|mother|father|mama|papa|bro|brother|sis|sister)\b', re.UNICODE | re.IGNORECASE)
    
    def transform(self, data):
        return ~data[self.column].str.extract(NameFamilial.NAIVE, expand=False).isnull()



import inspect
import warnings
from lore.models import base


class Base(base.Base):

    def __init__(self, *args, **kwargs):
        frame, filename, line_number, function_name, lines, index = inspect.stack()[1]
        warnings.showwarning('lore.models.Base does not fully support lore.estimators.keras.Base. You should inherit from a specific Base from available modules: lore.models.[keras|sklearn|xgboost].Base', DeprecationWarning,
                             filename, line_number)
        super(Base, self).__init__(*args, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging
import sys

from lore import env, util, ansi
from lore.ansi import underline
from lore.util import timer

logger = logging.getLogger(__name__)


__author__ = 'Montana Low and Jeremy Stanley'
__copyright__ = 'Copyright © 2017, Instacart'
__credits__ = ['Montana Low', 'Jeremy Stanley', 'Emmanuel Turlay']
__license__ = 'MIT'
__version__ = '0.5.0rc1'
__maintainer__ = 'Montana Low'
__email__ = 'montana@instacart.com'
__status__ = 'Development Status :: 3 - Alpha'


def banner():
    import socket
    import getpass
    
    return '%s in %s on %s' % (
        ansi.foreground(ansi.GREEN, env.project),
        ansi.foreground(env.color, env.name),
        ansi.foreground(ansi.CYAN,
                        getpass.getuser() + '@' + socket.gethostname())
    )


lore_no_env = False
if hasattr(sys, 'lore_no_env'):
    lore_no_env = sys.lore_no_env

if len(sys.argv) > 1 and sys.argv[0][-4:] == 'lore' and sys.argv[1] in ['install', 'init']:
    lore_no_env = True

if not lore_no_env:
    # everyone else gets validated and launched on import
    env.validate()
    env.launch()

if env.launched():
    print(banner())
    logger.info(banner())
    logger.debug('python environment: %s' % env.prefix)

    if not lore_no_env:
        with timer('check requirements', logging.DEBUG):
            env.check_requirements()



from abc import ABCMeta, abstractmethod

import hashlib
import inspect


class Base(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def __getitem__(self, key):
        pass

    @abstractmethod
    def __setitem__(self, key, value):
        pass

    @abstractmethod
    def __delitem__(self, key):
        pass

    @abstractmethod
    def __contains__(self, key):
        pass

    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def keys(self):
        pass

    @abstractmethod
    def values(self):
        pass

    @abstractmethod
    def batch_set(self):
        pass

    @abstractmethod
    def batch_get(self, data_dict):
        pass

    def key(self, *args, **kwargs):
        stack = inspect.stack()
        caller = kwargs.pop('caller', stack[-2])
        instance = kwargs.pop('instance', self)

        return '.'.join((
            instance.__module__,
            instance.__class__.__name__,
            caller.__code__.co_name,
            hashlib.sha1(str(args).encode('utf-8') + str(kwargs).encode('utf-8')).hexdigest()
        ))



from lore.estimators import keras, xgboost, sklearn
import warnings, inspect

class Keras(keras.Keras):

    def __init__(self, **kwargs):
        frame, filename, line_number, function_name, lines, index = inspect.stack()[1]
        warnings.showwarning('Please import Keras with "from lore.estimators.keras import Keras"', DeprecationWarning,
                             filename, line_number)
        super(Keras, self).__init__(**kwargs)

class XGBoost(xgboost.XGBoost):

    def __init__(self, **kwargs):
        frame, filename, line_number, function_name, lines, index = inspect.stack()[1]
        warnings.showwarning('Please import XGBoost with "from lore.estimators.xgboost import XGBoost"', DeprecationWarning,
                             filename, line_number)
        super(XGBoost, self).__init__(**kwargs)
 
class SKLearn(sklearn.SKLearn):

    def __init__(self, estimator):
        frame, filename, line_number, function_name, lines, index = inspect.stack()[1]
        warnings.showwarning('Please import SKLearn with "from lore.estimators.sklearn import SKLearn"', DeprecationWarning,
                             filename, line_number)
        super(SKLearn, self).__init__(estimator)




from __future__ import absolute_import

from abc import ABCMeta, abstractmethod
import codecs
from collections import namedtuple
import gc
import hashlib
import inspect
import io
import logging
import os
import pickle
import sqlite3
import types
import warnings

import numpy
import pandas
from sklearn.model_selection import train_test_split

from lore.util import timer, timed
import lore.stores


logger = logging.getLogger(__name__)


Observations = namedtuple('Observations', 'x y')


class Holdout(object):
    __metaclass__ = ABCMeta
    
    test_size = 0.1

    def __init__(self):
        self.name = self.__module__ + '.' + self.__class__.__name__
        self.stratify = None
        self.subsample = None
        self.split_seed = 1
        self._data = None
        self._encoders = None
        self._training_data = None
        self._test_data = None
        self._validation_data = None
        self._output_encoder = None
        self._encoded_training_data = None
        self._encoded_validation_data = None
        self._encoded_test_data = None

    def __getstate__(self):
        state = dict(self.__dict__)
        # bloat can be restored via self.__init__() + self.build()
        for bloat in [
            '_data',
            '_training_data',
            '_test_data',
            '_validation_data',
            '_encoded_training_data',
            '_encoded_validation_data'
            '_encoded_test_data',
        ]:
            state[bloat] = None
        return state

    @abstractmethod
    def get_data(self):
        pass

    @abstractmethod
    def get_encoders(self):
        pass

    @abstractmethod
    def get_output_encoder(self):
        pass

    @property
    def encoders(self):
        if self._encoders is None:
            with timer('fit encoders:'):
                self._encoders = self.get_encoders()
                for encoder in self._encoders:
                    encoder.fit(self.training_data)
            
        return self._encoders
    
    @property
    def output_encoder(self):
        if self._output_encoder is None:
            with timer('fit output encoder:'):
                self._output_encoder = self.get_output_encoder()
                self._output_encoder.fit(self.training_data)
            
        return self._output_encoder
    
    @property
    def training_data(self):
        if self._training_data is None:
            self._split_data()
    
        return self._training_data

    @property
    def validation_data(self):
        if self._validation_data is None:
            self._split_data()
    
        return self._validation_data

    @property
    def test_data(self):
        if self._test_data is None:
            self._split_data()
    
        return self._test_data

    @property
    def encoded_training_data(self):
        if not self._encoded_training_data:
            with timer('encode training data:'):
                self._encoded_training_data = self.observations(self.training_data)
    
        return self._encoded_training_data

    @property
    def encoded_validation_data(self):
        if not self._encoded_validation_data:
            with timer('encode validation data:'):
                self._encoded_validation_data = self.observations(self.validation_data)
    
        return self._encoded_validation_data

    @property
    def encoded_test_data(self):
        if not self._encoded_test_data:
            with timer('encode test data:'):
                self._encoded_test_data = self.observations(self.test_data)
    
        return self._encoded_test_data

    def observations(self, data):
        return Observations(x=self.encode_x(data), y=self.encode_y(data))

    @timed(logging.INFO)
    def encode_x(self, data):
        """
        :param data: unencoded input dataframe
        :return: a dict with encoded values
        """
        encoded = {}
        for encoder in self.encoders:
            transformed = encoder.transform(data)
            if hasattr(encoder, 'sequence_length'):
                for i in range(encoder.sequence_length):
                    encoded[encoder.sequence_name(i)] = transformed[:,i]
            else:
                encoded[encoder.name] = transformed
        
        # Using a DataFrame as a container temporairily requires double the memory,
        # as pandas copies all data on __init__. This is justified by having a
        # type supported by all dependent libraries (heterogeneous dict is not)
        return pandas.DataFrame(encoded)
    
    @timed(logging.INFO)
    def encode_y(self, data):
        return self.output_encoder.transform(data)

    @timed(logging.INFO)
    def decode(self, predictions):
        return {encoder.name: encoder.reverse_transform(predictions) for encoder in self.encoder}

    @timed(logging.INFO)
    def _split_data(self):
        if self._data:
            return

        numpy.random.seed(self.split_seed)
        logger.debug('random seed set to: %i' % self.split_seed)

        self._data = self.get_data()

        if self.subsample:

            if self.stratify:
                logger.debug('subsampling stratified by `%s`: %s' % (
                    self.stratify, self.subsample))
                ids = self._data[[self.stratify]].drop_duplicates()
                ids = ids.sample(self.subsample)
                self._data = pandas.merge(self._data, ids, on=self.stratify)
            else:
                logger.debug('subsampling rows: %s' % self.subsample)
                self._data = self._data.sample(self.subsample)
            gc.collect()

        if self.stratify:
            ids = self._data[self.stratify].drop_duplicates()

            train_ids, validate_ids = train_test_split(
                ids,
                test_size=self.test_size,
                random_state=1
            )
            train_ids, test_ids = train_test_split(
                train_ids,
                test_size=self.test_size,
                random_state=1
            )
        
            rows = self._data[self.stratify].values
            self._training_data = self._data.iloc[numpy.isin(rows, train_ids.values)]
            self._validation_data = self._data.iloc[numpy.isin(rows, validate_ids.values)]
            self._test_data = self._data.iloc[numpy.isin(rows, test_ids.values)]
        else:
            self._training_data, self._validation_data = train_test_split(
                self._data,
                test_size=self.test_size,
                random_state=1
            )
            
            self._training_data, self._test_data = train_test_split(
                self._training_data,
                test_size=self.test_size,
                random_state=1
            )
        # It's import to reset these indexes after split so in case
        # these dataframes are copied, the missing split rows are
        # not re-materialized later full of nans.
        self._training_data.reset_index(drop=True, inplace=True)
        self._validation_data.reset_index(drop=True, inplace=True)
        self._test_data.reset_index(drop=True, inplace=True)

        logger.debug('data: %i | training: %i | validation: %i | test: %i' % (
            len(self._data),
            len(self._training_data),
            len(self._validation_data),
            len(self._test_data)
        ))


class TimeSeries(Holdout):
    __metaclass__ = ABCMeta

    def __init__(self, test_size = 0.1, sort_by = None):
        super(TimeSeries, self).__init__()
        self.sort_by = sort_by
        self.test_size = test_size

    @timed(logging.INFO)
    def _split_data(self):
        if self._data:
            return

        logger.debug('No shuffle test train split')

        self._data = self.get_data()

        if self.sort_by:
            self._data = self._data.sort_values(by=self.sort_by, ascending=True)
        test_rows = int(len(self._data) * self.test_size)
        valid_rows = test_rows
        train_rows = int(len(self._data) - test_rows - valid_rows)
        self._training_data = self._data.iloc[:train_rows]
        self._validation_data = self._data[train_rows:train_rows+valid_rows]
        self._test_data = self._data.iloc[-test_rows:]


class LowMemory(Holdout):
    connection = sqlite3.connect(os.path.join(lore.env.data_dir, 'low_memory.sqlite'), isolation_level=None)
    connection.execute('PRAGMA synchronous = OFF')
    connection.execute('PRAGMA journal_mode = MEMORY')

    @timed(logging.INFO)
    def __init__(self):
        super(LowMemory, self).__init__()
        self._training_key = None
        self._validation_key = None
        self._test_key = None
        self._table = None
        self.chunksize = None
        self.loaded = False
        self._columns = None
        self._length = None
        self._datetime_columns = None
    
    @property
    def table(self):
        if self._table is None:
            self._table = 'pipeline_' + self.__class__.__name__.lower() + '_' + hashlib.sha1(str(self.__dict__).encode('utf-8')).hexdigest()
            
        return self._table
    
    @property
    def table_training(self):
        return self.table + '_training'
    
    @property
    def table_validation(self):
        return self.table + '_validation'
    
    @property
    def table_test(self):
        return self.table + '_test'

    @property
    def training_data(self):
        return self.read_table(self.table_training)
        
    @property
    def validation_data(self):
        return self.read_table(self.table_validation)

    @property
    def test_data(self):
        return self.read_table(self.table_test)
        
    def read_table(self, name):
        if not self.loaded:
            self._split_data()
    
        return pandas.read_sql(
            "SELECT * FROM {name}".format(name=name),
            self.connection,
            chunksize=self.chunksize,
            parse_dates=self.datetime_columns
        )

    @property
    def datetime_columns(self):
        if self._datetime_columns is None:
            with open(self.metadata_path, 'rb') as f:
                self._datetime_columns = pickle.load(f)
                
        return self._datetime_columns

    @property
    def encoded_training_data(self):
        if not self._encoded_training_data:
            if not self.loaded:
                self._split_data()
            with timer('encode training data:'):
                self._encoded_training_data = self.observations(self.table_training)
    
        return self._encoded_training_data

    @property
    def encoded_validation_data(self):
        if not self._encoded_validation_data:
            if not self.loaded:
                self._split_data()
            with timer('encode validation data:'):
                self._encoded_validation_data = self.observations(self.table_validation)
    
        return self._encoded_validation_data

    @property
    def encoded_test_data(self):
        if not self._encoded_test_data:
            if not self.loaded:
                self._split_data()
            with timer('encode test data:'):
                self._encoded_test_data = self.observations(self.table_test)
    
        return self._encoded_test_data

    @property
    def encoders(self):
        if self._encoders is None:
            with timer('fit encoders:'):
                self._encoders = self.get_encoders()
                for encoder in self._encoders:
                    encoder.fit(self.read_column(self.table_training, encoder.source_column))
                    gc.collect()

        return self._encoders
    
    def read_column(self, table, column):
        return pandas.read_sql(
            'SELECT {column} FROM {table}'.format(
                column=self.quote(column),
                table=self.quote(table)
            ),
            self.connection,
            parse_dates=set(self.datetime_columns).intersection([column])
        )
    
    @property
    def output_encoder(self):
        if self._output_encoder is None:
            with timer('fit output encoder:'):
                self._output_encoder = self.get_output_encoder()
                self._output_encoder.fit(self.read_column(self.table_training, self._output_encoder.source_column))
        
        return self._output_encoder
    
    def observations(self, data):
        return Observations(x=self.encode_x(data), y=self.encode_y(data))
    
    @timed(logging.INFO)
    def encode_x(self, data):
        if isinstance(data, pandas.DataFrame):
            return super(LowMemory, self).encode_x(data)
        
        encoded = {}
        for encoder in self.encoders:
            transformed = encoder.transform(self.read_column(data, encoder.source_column))
            if hasattr(encoder, 'sequence_length'):
                for i in range(encoder.sequence_length):
                    encoded[encoder.sequence_name(i)] = transformed[:, i]
            else:
                encoded[encoder.name] = transformed
        
        # Using a DataFrame as a container temporairily requires double the memory,
        # as pandas copies all data on __init__. This is justified by having a
        # type supported by all dependent libraries (heterogeneous dict is not)
        return pandas.DataFrame(encoded)
    
    @timed(logging.INFO)
    def encode_y(self, data):
        if isinstance(data, pandas.DataFrame):
            return super(LowMemory, self).encode_x(data)
            
        return self.output_encoder.transform(self.read_column(data, self._output_encoder.source_column))
    
    @timed(logging.INFO)
    def decode(self, predictions):
        return {encoder.name: encoder.reverse_transform(predictions) for encoder in self.encoder}
    
    @property
    def metadata_path(self):
        return os.path.join(lore.env.data_dir,  self.table + '.pickle')
    
    def _split_data(self):
        self.loaded = True
    
        if self.connection.execute(
            """
                SELECT name
                FROM sqlite_master
                WHERE type='table'
                  AND name={name}
            """.format(
                name=self.quote(self.table)
            )
        ).fetchone():
            return
    
        self._data = self.get_data()
    
        if not isinstance(self._data, types.GeneratorType):
            raise TypeError('LowMemory pipelines must be passed a generator rather than a dataframe. Did you forget to pass a `chunksize` to lore.io.Connection.dataframe()?')
    
        self.length = 0

        # This loop is unfortunately CPU limited rather than IO bound, based on dataframe creation
        for i, dataframe in enumerate(self._data):
            if i == 0:
                self._columns = dataframe.columns
                self.chunksize = len(dataframe)
                buffer = io.StringIO()
                dataframe.info(buf=buffer, memory_usage='deep')
                logger.info(buffer.getvalue())
                logger.info(dataframe.head())

                self._datetime_columns = [col for col, type in dataframe.dtypes.iteritems() if type == 'datetime64[ns]']
                with open(self.metadata_path, 'wb') as f:
                    pickle.dump(self.datetime_columns, f, pickle.HIGHEST_PROTOCOL)

            logger.info('appending %i rows to sqlite' % len(dataframe))
            self.length += len(dataframe)
            dataframe.to_sql(self.table, self.connection, index=False, if_exists="append")
    
        if self.subsample:
            self.connection.executescript(
                """
                    BEGIN;
                    
                    CREATE TABLE {subsample} AS
                    SELECT * FROM {table}
                    ORDER BY random()
                    LIMIT {limit};
                    
                    DROP TABLE {table};
                    
                    ALTER TABLE {subsample} RENAME TO {table};
                    
                    COMMIT;
                """.format(
                    table=self.quote(self.table),
                    subsample=self.quote(self.table + '_subsample'),
                    limit=self.subsample
                )
            )
            self.length = self.subsample

        self._random_split(self.table_training, 0, 0.8, stratify=self.stratify)
        self._random_split(self.table_validation, 0.8, 0.9, stratify=self.stratify)
        self._random_split(self.table_test, 0.9, 1, stratify=self.stratify)
        
        logger.debug('data: %i | training: %i | validation: %i | test: %i' % (
            self.length,
            self.table_length(self.table_training),
            self.table_length(self.table_validation),
            self.table_length(self.table_test),
        ))

    def __getitem__(self, column):
        return self.connection.execute(
            'SELECT {column} FROM {table}'.format(
                column=self.quote(column),
                table=self.quote(self.table),
            )
        )

    def __len__(self):
        if self._length is None:
            self._length = self.table_length(self.table)
        return self._length

    def table_length(self, name):
        return self.connection.execute(
            'SELECT count(_ROWID_) FROM {table}'.format(
                table=self.quote(name)
            )
        ).fetchone()[0]
    
    @property
    def columns(self):
        if not self.loaded:
            self._split_data()
    
        if self._columns is None:
            self._columns = [
                row[1] for row in self.connection.execute(
                    'PRAGMA table_info({table});'.format(table=self.quote(self.table))
                ).fetchall()
            ]
    
        return self._columns

    def _random_split(self, name, start, stop, stratify=None):
        random = self.quote(name + '_random')
        name = self.quote(name)
        
        if stratify is not None:
            self.connection.execute("""
                CREATE TABLE IF NOT EXISTS {random} AS
                SELECT DISTINCT {stratify}
                FROM {source}
                ORDER BY random();
            """.format(
                random=random,
                source=self.quote(self.table),
                stratify=self.quote(stratify)
            ))
            random_column = self.quote(stratify)
    
        else:
            self.connection.execute("""
                CREATE TABLE IF NOT EXISTS {random} AS
                SELECT _rowid_ as other_rowid
                FROM {source}
                ORDER BY random();
            """.format(
                random=random,
                source=self.quote(self.table)
            ))
            stratify = '_rowid_'
            random_column = 'other_rowid'
    
        self.connection.execute("""
            CREATE TABLE IF NOT EXISTS {split} AS
            SELECT {source}.*
            FROM {source}
            JOIN {random}
              ON {random}.{random_column} = {source}.{stratify}
            WHERE {random}._rowid_ - 1 >= (select max(_rowid_) from {random}) * {start}
              AND {random}._rowid_ - 1 < (select max(_rowid_) from {random}) * {stop}
        """.format(
            split=name,
            source=self.quote(self.table),
            random=random,
            random_column=random_column,
            start=start,
            stop=stop,
            stratify=stratify
        ))

    def quote(self, identifier, errors="strict"):
        """
        https://stackoverflow.com/questions/6514274/how-do-you-escape-strings-for-sqlite-table-column-names-in-python

        :param identifier:
        :param errors:
        :return:
        """
        encodable = identifier.encode("utf-8", errors).decode("utf-8")
    
        nul_index = encodable.find("\x00")
    
        if nul_index >= 0:
            error = UnicodeEncodeError(
                "NUL-terminated utf-8",
                encodable,
                nul_index,
                nul_index + 1,
                "NUL not allowed"
            )
            error_handler = codecs.lookup_error(errors)
            replacement, _ = error_handler(error)
            encodable = encodable.replace("\x00", replacement)
    
        return "\"" + encodable.replace("\"", "\"\"") + "\""


class TrainTestSplit(Holdout):
    def __init__(self, **kwargs):
        frame, filename, line_number, function_name, lines, index = inspect.stack()[1]
        warnings.showwarning(
            'TrainTestSplit has been renamed to Holdout. Please update your code.',
            DeprecationWarning,
            filename,
            line_number
        )
        super(TrainTestSplit, self).__init__(**kwargs)


class SortedTrainTestSplit(TimeSeries):
    def __init__(self, **kwargs):
        frame, filename, line_number, function_name, lines, index = inspect.stack()[1]
        warnings.showwarning(
            'SortedTrainTestSplit has been renamed to TimeSeries. Please update your code.',
            DeprecationWarning,
            filename,
            line_number
        )
        super(SortedTrainTestSplit, self).__init__(**kwargs)



import logging
from datetime import datetime

import keras.callbacks

from lore.util import timer


logger = logging.getLogger(__name__)


class ReloadBest(keras.callbacks.ModelCheckpoint):
    def __init__(
        self,
        filepath,
        monitor='val_loss',
        mode='auto',
    ):
        super(ReloadBest, self).__init__(
            filepath=filepath,
            monitor=monitor,
            verbose=0,
            mode=mode,
            save_best_only=False,
            save_weights_only=True,
            period=1
        )
        self.train_loss = None
        self.validate_loss = None
        self.best_epoch = None
        self.train_begin = None

    def on_train_begin(self, logs=None):
        super(ReloadBest, self).on_train_begin(logs)

        self.train_begin = datetime.utcnow()
        logger.info('=============================================')
        logger.info('|    epoch |     time |    train | validate |')
        logger.info('---------------------------------------------')
    
    def on_train_end(self, logs=None):
        super(ReloadBest, self).on_train_end(logs)
        logger.info('=============================================')
        if self.best_epoch is not None:
            logger.debug('best epoch: %i' % self.best_epoch)
            with timer('load best epoch:'):
                self.model.load_weights(
                    self.filepath.format(epoch=self.best_epoch)
                )
    
    def on_epoch_end(self, epoch, logs=None):
        super(ReloadBest, self).on_epoch_end(epoch, logs)
        time = datetime.utcnow() - self.train_begin
        train_loss = logs.get('loss')
        validate_loss = logs.get('val_loss')
        if validate_loss:
            if self.validate_loss is None or self.validate_loss > validate_loss:
                self.best_epoch = epoch + 1
                self.train_loss = train_loss
                self.validate_loss = validate_loss
        else:
            logger.error('No val_loss in logs, setting to NaN')
            validate_loss = float('nan')
        logger.info('| %8i | %8s | %8.4f | %8.4f |' % (
            epoch, str(time).split('.', 2)[0], train_loss, validate_loss)
        )



from __future__ import absolute_import

import inspect
import json
import logging
import os.path
from os.path import join
import pickle
import re

from tabulate import tabulate
import lore.estimators
import lore.serializers
from lore.util import timer

from sklearn.model_selection import RandomizedSearchCV

logger = logging.getLogger(__name__)


class Base(object):
    def __init__(self, pipeline=None, estimator=None):
        self.name = self.__module__ + '.' + self.__class__.__name__
        self._estimator = None
        self.estimator = estimator
        self.fitting = None
        self.pipeline = pipeline
    
    def __getstate__(self):
        return dict(self.__dict__)
    
    @property
    def estimator(self):
        return self._estimator
    
    @estimator.setter
    def estimator(self, value):
        self._estimator = value
        
        # Keras models require access to the pipeline during build,
        # and the serializer during fit for extended functionality
        if hasattr(self._estimator, 'model'):
            self._estimator.model = self

    def fit(self, **estimator_kwargs):
        self.fitting = self.__class__.last_fitting() + 1

        self.stats = self.estimator.fit(
            x=self.pipeline.encoded_training_data.x,
            y=self.pipeline.encoded_training_data.y,
            **estimator_kwargs
        )
        self.save(stats=self.stats)
        logger.info(
            '\n\n' + tabulate([self.stats.keys(), self.stats.values()], tablefmt="grid", headers='firstrow') + '\n\n')
    
    def predict(self, dataframe):
        return self.estimator.predict(self.pipeline.encode_x(dataframe))
    
    def hyper_parameter_search(
            self,
            param_distributions,
            n_iter=10,
            scoring=None,
            fit_params={},
            n_jobs=1,
            iid=True,
            refit=True,
            cv=None,
            verbose=0,
            pre_dispatch='2*njobs',
            random_state=None,
            error_score='raise',
            return_train_score=True
    ):
        """Random search hyper params
        """
        if scoring is None:
            scoring = None
        result = RandomizedSearchCV(
            self.estimator,
            param_distributions,
            n_iter=n_iter,
            scoring=scoring,
            n_jobs=n_jobs,
            iid=iid,
            refit=refit,
            cv=cv,
            verbose=verbose,
            pre_dispatch=pre_dispatch,
            random_state=random_state,
            error_score=error_score,
            return_train_score=return_train_score
        ).fit(
            self.pipeline.encoded_training_data.x,
            y=self.pipeline.encoded_training_data.y,
            **fit_params
        )
        self.estimator = result.best_estimator_
        
        return result
    
    @classmethod
    def local_path(cls):
        return join(lore.env.models_dir, cls.remote_path())
    
    @classmethod
    def remote_path(cls):
        return join(cls.__module__, cls.__name__)
    
    @classmethod
    def last_fitting(cls):
        if not os.path.exists(cls.local_path()):
            return 0
        
        fittings = [int(d) for d in os.listdir(cls.local_path()) if re.match(r'^\d+$', d)]
        if not fittings:
            return 0
        
        return sorted(fittings)[-1]
    
    def fitting_path(self):
        if self.fitting is None:
            self.fitting = self.last_fitting()
        
        return join(self.local_path(), str(self.fitting))
    
    def model_path(self):
        return join(self.fitting_path(), 'model.pickle')
    
    def remote_model_path(self):
        return join(self.remote_path(), 'model.pickle')
    
    def save(self, stats=None):
        if self.fitting is None:
            raise ValueError("This model has not been fit yet. There is no point in saving.")

        if not os.path.exists(self.fitting_path()):
            os.makedirs(self.fitting_path())

        with timer('pickle model:'):
            with open(self.model_path(), 'wb') as f:
                pickle.dump(self, f)
        
        with open(join(self.fitting_path(), 'params.json'), 'w') as f:
            params = {}
            for child in [self.estimator, self.pipeline]:
                param = child.__module__ + '.' + child.__class__.__name__
                params[param] = {}
                for key, value in child.__getstate__().items():
                    if not key.startswith('_'):
                        params[param][key] = value.__repr__()
            json.dump(params, f, indent=2, sort_keys=True)
        
        if stats:
            with open(join(self.fitting_path(), 'stats.json'), 'w') as f:
                json.dump(stats, f, indent=2, sort_keys=True)

    @classmethod
    def load(cls, fitting=None):
        model = cls()
        if fitting is None:
            model.fitting = model.last_fitting()
        else:
            model.fitting = fitting
        
        with timer('unpickle model:'):
            with open(model.model_path(), 'rb') as f:
                loaded = pickle.load(f)
                loaded.fitting = model.fitting
                return loaded
    
    def upload(self):
        self.fitting = 0
        self.save()
        lore.io.upload(self.model_path(), self.remote_model_path())
     
    @classmethod
    def download(cls, fitting=0):
        model = cls(None, None)
        model.fitting = fitting
        lore.io.download(model.model_path(), model.remote_model_path())
        return cls.load(fitting)



from __future__ import absolute_import
import logging
from sklearn.base import BaseEstimator
from lore.util import timed


class SKLearn(BaseEstimator):
    def __init__(self, estimator):
        super(SKLearn, self).__init__()
        self.sklearn = estimator

    @timed(logging.INFO)
    def fit(self, x, y,  **sklearn_kwargs):
        self.sklearn.fit(x, y=y, **sklearn_kwargs)

        # TODO interesting SKLearn fitting stats
        return {}

    @timed(logging.INFO)
    def predict(self, dataframe):
        return self.sklearn.predict(dataframe)



import hashlib
import inspect
import logging
import io
import os
import re
import sys
import tempfile
import csv
import gzip
from datetime import datetime

from sqlalchemy import event
from sqlalchemy.engine import Engine
from sqlalchemy.schema import DropTable
from sqlalchemy.ext.compiler import compiles

import pandas
import sqlalchemy

import lore
from lore.util import timer
from lore.stores import query_cached


logger = logging.getLogger(__name__)


@compiles(DropTable, 'postgresql')
def _compile_drop_table(element, compiler, **kwargs):
    return compiler.visit_drop_table(element) + ' CASCADE'


_after_replace_callbacks = []
def after_replace(func):
    global _after_replace_callbacks
    _after_replace_callbacks.append(func)


@event.listens_for(Engine, "before_cursor_execute", retval=True)
def comment_sql_calls(conn, cursor, statement, parameters, context, executemany):
    conn.info.setdefault('query_start_time', []).append(datetime.now())
    
    stack = inspect.stack()[1:-1]
    if sys.version_info.major == 3:
        stack = [(x.filename, x.lineno, x.function) for x in stack]
    else:
        stack = [(x[1], x[2], x[3]) for x in stack]
    
    paths = [x[0] for x in stack]
    origin = next((x for x in paths if lore.env.project in x), None)
    if origin is None:
        origin = next((x for x in paths if 'sqlalchemy' not in x), None)
    if origin is None:
        origin = paths[0]
    caller = next(x for x in stack if x[0] == origin)
    
    statement = "/* %s | %s:%d in %s */\n" % (lore.env.project, caller[0], caller[1], caller[2]) + statement
    logger.debug(statement)
    return statement, parameters


@event.listens_for(Engine, "after_cursor_execute")
def time_sql_calls(conn, cursor, statement, parameters, context, executemany):
    total = datetime.now() - conn.info['query_start_time'].pop(-1)
    logger.info("SQL: %s" % total)


class Connection(object):
    UNLOAD_PREFIX = os.path.join(lore.env.name, 'unloads')
    IAM_ROLE = os.environ.get('IAM_ROLE', None)
    
    def __init__(self, url, name='connection', **kwargs):
        for int_value in ['pool_size', 'pool_recycle', 'max_overflow']:
            if int_value in kwargs:
                kwargs[int_value] = int(kwargs[int_value])
        if 'poolclass' in kwargs:
            kwargs['poolclass'] = getattr(sqlalchemy.pool, kwargs['poolclass'])
        if '__name__' in kwargs:
            del kwargs['__name__']
        if 'isolation_level' not in kwargs:
            kwargs['isolation_level'] = 'AUTOCOMMIT'

        self._engine = sqlalchemy.create_engine(url, **kwargs)
        self._connection = None
        self._metadata = None
        self.name = name
        self._transactions = []

    def __enter__(self):
        if self._connection is None:
            self._connection = self._engine.connect()
        dbapi_connection = self._connection.connection.connection
        if self._transactions:
            assert not dbapi_connection.autocommit
        else:
            assert dbapi_connection.autocommit
            dbapi_connection.autocommit = False
        self._transactions.append(self._connection.begin())
        return self
    
    def __exit__(self, type, value, traceback):
        transaction = self._transactions.pop()
        if type is None:
            transaction.commit()
        else:
            transaction.rollback()
        if not self._transactions:
            dbapi_connection = self._connection.connection.connection
            assert not dbapi_connection.autocommit
            dbapi_connection.autocommit = True

    @staticmethod
    def path(filename, extension='.sql'):
        return os.path.join(
            lore.env.root, lore.env.project, 'extracts',
            filename + extension)

    def execute(self, sql=None, filename=None, **kwargs):
        self.__execute(self.__prepare(sql, filename), kwargs)

    def insert(self, table, dataframe, batch_size=None):
        if batch_size is None:
            batch_size = len(dataframe)

        if self._connection is None:
            self._connection = self._engine.connect()

        if self._engine.dialect.name in ['postgresql', 'redshift']:
            if sys.version_info[0] == 2:
                rows = io.BytesIO()
            else:
                rows = io.StringIO()
            dataframe.to_csv(rows, index=False, header=False, sep='|', quoting=csv.QUOTE_NONE)
            rows.seek(0)
            self._connection.connection.cursor().copy_from(rows, table, sep='|', columns=dataframe.columns)
        else:
            dataframe.to_sql(
                table,
                self._connection,
                if_exists='append',
                index=False,
                chunksize=batch_size
            )

    def close(self):
        self._engine.dispose()
        self._connection = None
        
    def replace(self, table, dataframe, batch_size=None):
        import migrate.changeset
        global _after_replace_callbacks
        
        with timer('REPLACE ' + table):
            suffix = datetime.now().strftime('_%Y%m%d%H%M%S').encode('utf-8')
            self.metadata
            temp = 'tmp_'.encode('utf-8')
            source = sqlalchemy.Table(table, self.metadata, autoload=True, autoload_with=self._engine)
            destination_name = 'tmp_' + hashlib.sha256(temp + table.encode('utf-8') + suffix).hexdigest()[0:56]
            destination = sqlalchemy.Table(destination_name, self.metadata, autoload=False)
            for column in source.columns:
                destination.append_column(column.copy())
            destination.create()

            original_names = {}
            for index in source.indexes:
                # make sure the name is < 63 chars with the suffix
                name = hashlib.sha256(temp + index.name.encode('utf-8') + suffix).hexdigest()[0:60]
                original_names[name] = index.name
                columns = []
                for column in index.columns:
                    columns.append(next(x for x in destination.columns if x.name == column.name))
                new = sqlalchemy.Index(name, *columns)
                new.unique = index.unique
                new.table = destination
                new.create(bind=self._connection)
            self.insert(destination.name, dataframe, batch_size=batch_size)
            self.execute("BEGIN; SET LOCAL statement_timeout = '1min'; ANALYZE %s; COMMIT;" % self.quote_identifier(table))

            with self as transaction:
                backup = sqlalchemy.Table(table + '_b', self.metadata)
                backup.drop(bind=self._connection, checkfirst=True)
                source.rename(name=source.name + '_b', connection=self._connection)
                destination.rename(name=table, connection=self._connection)
                for index in source.indexes:
                    index.rename(index.name[0:-2] + '_b', connection=self._connection)
                for index in destination.indexes:
                    index.rename(original_names[index.name], connection=self._connection)
        
        for func in _after_replace_callbacks:
            func(destination, source)
        
    @property
    def metadata(self):
        if not self._metadata:
            self._metadata = sqlalchemy.MetaData(bind=self._engine)

        return self._metadata

    def select(self, sql=None, filename=None, **kwargs):
        cache = kwargs.pop('cache', False)
        sql = self.__prepare(sql, filename)
        return self._select(sql, kwargs, cache=cache)

    @query_cached
    def _select(self, sql, bindings):
        return self.__execute(sql, bindings).fetchall()

    def unload(self, sql=None, filename=None, **kwargs):
        cache = kwargs.pop('cache', False)
        sql = self.__prepare(sql, filename)
        return self._unload(sql, kwargs, cache=cache)
    
    @query_cached
    def _unload(self, sql, bindings):
        key = hashlib.sha1(str(sql).encode('utf-8')).hexdigest()

        match = re.match(r'.*?select\s(.*)from.*', sql, flags=re.IGNORECASE | re.UNICODE | re.DOTALL)
        if match:
            columns = []
            nested = 0
            potential = match[1].split(',')
            for column in potential:
                nested += column.count('(')
                nested -= column.count(')')
                if nested == 0:
                    columns.append(column.split()[-1].split('.')[-1].strip())
                elif column == potential[-1]:
                    column = re.split('from', column, flags=re.IGNORECASE)[0].strip()
                    columns.append(column.split()[-1].split('.')[-1].strip())
        else:
            columns = []
        logger.warning("Redshift unload requires poorly parsing column names from sql, found: {}".format(columns))

        sql = "UNLOAD ('" + sql.replace('\\', '\\\\').replace("'", "\\'") + "') "
        sql += "TO 's3://" + os.path.join(
            lore.io.bucket.name,
            self.UNLOAD_PREFIX,
            key,
            ''
        ) + "' "
        if Connection.IAM_ROLE:
            sql += "IAM_ROLE '" + Connection.IAM_ROLE + "' "
        sql += "DELIMITER '|' ADDQUOTES GZIP ALLOWOVERWRITE"
        if re.match(r'(.*?)(limit\s+\d+)(.*)', sql, re.IGNORECASE | re.UNICODE | re.DOTALL):
            logger.warning('LIMIT clause is not supported by unload, returning full set.')
            sql = re.sub(r'(.*?)(limit\s+\d+)(.*)', r'\1\3', sql, flags=re.IGNORECASE | re.UNICODE | re.DOTALL)
        self.__execute(sql, bindings)
        return key, columns

    @query_cached
    def load(self, key, columns):
        result = [columns]
        with timer('load:'):
            for entry in lore.io.bucket.objects.filter(
                Prefix=os.path.join(self.UNLOAD_PREFIX, key)
            ):
                temp = tempfile.NamedTemporaryFile()
                lore.io.bucket.download_file(entry.key, temp.name)
                with gzip.open(temp.name, 'rt') as gz:
                    result += list(csv.reader(gz, delimiter='|', quotechar='"'))
        
            return result
    
    @query_cached
    def load_dataframe(self, key, columns):
        with timer('load_dataframe:'):
            frames = []
            for entry in lore.io.bucket.objects.filter(
                Prefix=os.path.join(self.UNLOAD_PREFIX, key)
            ):
                temp = tempfile.NamedTemporaryFile()
                lore.io.bucket.download_file(entry.key, temp.name)
                dataframe = pandas.read_csv(
                    temp.name,
                    delimiter='|',
                    quotechar='"',
                    compression='gzip',
                    error_bad_lines=False
                )
                dataframe.columns = columns
                frames.append(dataframe)

            result = pandas.concat(frames)
            result.columns = columns
            buffer = io.StringIO()
            result.info(buf=buffer, memory_usage='deep')
            logger.info(buffer.getvalue())
            logger.info(result.head())
            return result
        
    def dataframe(self, sql=None, filename=None, **kwargs):
        cache = kwargs.pop('cache', False)
        chunksize = kwargs.pop('chunksize', None)
        if chunksize and cache:
            raise ValueError('Chunking is incompatible with caching. Choose to pass either "cache" or "chunksize".')
        sql = self.__prepare(sql, filename)
        dataframe = self._dataframe(sql, kwargs, cache=cache, chunksize=chunksize)
        if chunksize is None:
            buffer = io.StringIO()
            dataframe.info(buf=buffer, memory_usage='deep')
            logger.info(buffer.getvalue())
            logger.info(dataframe.head())
        return dataframe
        
    @query_cached
    def _dataframe(self, sql, bindings, chunksize=None):
        with timer("dataframe:"):
            if self._connection is None:
                self._connection = self._engine.connect()
            return pandas.read_sql(sql=sql, con=self._connection, params=bindings, chunksize=chunksize)

    def quote_identifier(self, identifier):
        return self._engine.dialect.identifier_preparer.quote(identifier)
        
    def __prepare(self, sql, filename):
        if sql is None and filename is not None:
            filename = Connection.path(filename, '.sql')
            logger.debug("READ SQL FILE: " + filename)
            with open(filename) as file:
                sql = file.read()
        # support mustache style bindings
        sql = re.sub(r'\{(\w+?)\}', r'%(\1)s', sql)
        return sql

    def __execute(self, sql, bindings):
        if self._connection is None:
            self._connection = self._engine.connect()
        return self._connection.execute(sql, bindings)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals

from builtins import input

import argparse
import glob
import os
import sys
import platform
import re
import shutil
import subprocess
import logging
from io import open

import lore
from lore import ansi, env, util
from lore.util import timer, which

logger = logging.getLogger(__name__)


class HelpfulParser(argparse.ArgumentParser):
    def error(self, message):
        self.print_help(sys.stderr)
        self.exit(2, '%s: error: %s\n' % (self.prog, message))

    
def main(args=None):
    parser = HelpfulParser(prog='lore')
    parser.add_argument('--version', action='version',
                        version='lore %s' % lore.__version__)

    commands = parser.add_subparsers( help='common commands')
    
    init_parser = commands.add_parser('init', help='create a new lore project')
    init_parser.set_defaults(func=init)
    init_parser.add_argument('NAME', help='the name of the project')
    init_parser.add_argument('--git-ignore', default=True)
    init_parser.add_argument('--python-version', default=None)

    api_parser = commands.add_parser(
        'api',
        help='serve the api'
    )
    api_parser.set_defaults(func=api)

    console_parser = commands.add_parser(
        'console',
        help='launch an interactive python shell'
    )
    console_parser.set_defaults(func=console)

    exec_parser = commands.add_parser(
        'exec',
        help='run a shell command in this project\'s virtual env'
    )
    exec_parser.set_defaults(func=execute)

    install_parser = commands.add_parser(
        'install',
        help='install dependencies in a virtualenv'
    )
    install_parser.set_defaults(func=install)
    install_parser.add_argument(
        '--native',
        help='build optimized native dependencies (tensorflow)',
        action='store_true'
    )
    install_parser.add_argument(
        '--upgrade',
        help='recalculate requirements.frozen.txt with current versions',
        action='store_true'
    )

    task_parser = commands.add_parser(
        'task',
        help='run a task from the command line'
    )
    task_parser.set_defaults(func=task)

    pip_parser = commands.add_parser(
        'pip',
        help='pass a command to this project\'s virtual env pip'
    )
    pip_parser.set_defaults(func=pip)

    python_parser = commands.add_parser(
        'python',
        help='pass a command to this project\'s virtual env python'
    )
    python_parser.set_defaults(func=python)

    python_parser = commands.add_parser(
        'notebook',
        help='pass a command to this project\'s virtual env jupyter notebook'
    )
    python_parser.set_defaults(func=notebook)

    test_parser = commands.add_parser(
        'test',
        help='run tests'
    )
    test_parser.add_argument(
        'modules',
        nargs='?',
        help='test only certain modules, e.g. tests.unit.test_foo,tests.unit.test_bar'
    )
    test_parser.set_defaults(func=test)

    (known, unknown) = parser.parse_known_args(args)
    if '--env-launched' in unknown:
        unknown.remove('--env-launched')
    if '--env-checked' in unknown:
        unknown.remove('--env-checked')
    known.func(known, unknown)
    

def api(parsed, unknown):
    api_path = os.path.join(env.root, env.project, 'api')
    endpoint_paths = []
    consumer_paths = []
    if 'HUB_LISTENERS' in os.environ:
        for name in os.environ.get('HUB_LISTENERS').split(','):
            endpoint = os.path.join(api_path, name + '_endpoint.py')
            consumer = os.path.join(api_path, name + '_consumer.py')
            if os.path.exists(endpoint):
                endpoint_paths.append(endpoint)
            elif os.path.exists(consumer):
                consumer_paths.append(consumer)
            else:
                raise IOError('No file found for listener "%s". The following paths were checked:\n  %s\n  %s' % (name, consumer, endpoint))
    else:
        endpoint_paths = glob.glob(os.path.join(api_path, '*_endpoint.py'))
        consumer_paths = glob.glob(os.path.join(api_path, '*_consumer.py'))
     
    for path in endpoint_paths + consumer_paths:
        module = os.path.basename(path)[:-3]
        if sys.version_info.major == 2:
            import imp
            imp.load_source(module, path)
        elif sys.version_info.major == 3:
            import importlib.util
            spec = importlib.util.spec_from_file_location(module, path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
    util.strip_one_off_handlers()

    if len(endpoint_paths) > 0 and len(consumer_paths) > 0:
        from hub.listeners.combined import CombinedListener as Listener
    elif len(endpoint_paths) > 0:
        from hub.listeners.endpoint import EndpointListener as Listener
    elif len(consumer_paths) > 0:
        from hub.listeners.consumer import ConsumerListener as Listener
    else:
        raise IOError('No hub listeners found in %s' % api_path)
    
    Listener(
        os.environ.get('HUB_APP_NAME', env.project),
        concurrency=os.environ.get("HUB_CONCURRENCY", 4),
        host_index = os.environ.get("RABBIT_HOST_INDEX")
    ).start()


def console(parsed, unknown):
    install_jupyter_kernel()
    sys.argv[0] = env.bin_jupyter
    args = [env.bin_jupyter, 'console', '--kernel', env.project] + unknown
    print(ansi.success('JUPYTER') + ' ' + str(env.bin_jupyter))
    os.execv(env.bin_jupyter, args)


def execute(parsed, unknown):
    if len(unknown) == 0:
        print(ansi.error() + ' no args to execute!')
        return
    
    print(ansi.success('EXECUTE ') + ' '.join(unknown))
    
    os.environ['PATH'] = os.path.join(env.prefix, 'bin') + ':' + os.environ['PATH']
    subprocess.check_call(unknown, env=os.environ)


def init(parsed, unknown):
    template = os.path.join(os.path.dirname(__file__), 'template')
    
    if os.path.exists(parsed.NAME):
        sys.exit(ansi.error() + ' "' + parsed.NAME + '" already exists in this directoy!')
        
    shutil.copytree(template, parsed.NAME, symlinks=False, ignore=None)
    os.chdir(parsed.NAME)
    shutil.move('app', parsed.NAME)

    with open('requirements.txt', 'wt') as file:
        file.write('lore\n')
    
    if parsed.python_version:
        python_version = parsed.python_version
    else:
        python_version = '3.6.2'

    with open('runtime.txt', 'wt') as file:
        file.write('python-' + python_version + '\n')
    if sys.version_info[0] == 2:
        reload(lore.env)
    else:
        import importlib
        importlib.reload(lore.env)
    install(parsed, unknown)


def install(parsed, unknown):
    env.validate()
    if platform.system() == 'Darwin':
        install_darwin()
    elif platform.system() == 'Linux':
        install_linux()
    elif platform.system() == 'Windows':
        print(
            ansi.warning() + ' pyenv does not '
                             'support Windows. You\'ll have the task of managing'
                             ' python versions yourself 😰'
        )
        return
    else:
        raise KeyError('unknown system: ' % platform.system())

    install_python_version()
    create_virtual_env()
    install_requirements(parsed)
    
    if hasattr(parsed, 'native') and parsed.native:
        install_tensorflow()


def task(parsed, unknown):
    task = unknown[0]
    print(ansi.error() + 'not implemented')
    
    
def pip(parsed, unknown):
    args = [env.bin_python, '-m', 'pip'] + unknown
    print(ansi.success('EXECUTE ') + ' '.join(args))
    subprocess.check_call(args)


def python(parsed, unknown):
    args = [env.bin_python] + unknown
    print(ansi.success('EXECUTE ') + ' '.join(args))
    subprocess.check_call(args)


def test(parsed, unknown):
    with timer('boot time'):
        if 'LORE_ENV' not in os.environ:
            env.name = env.TEST
            logger.level = logging.WARN
        
        import unittest
        if parsed.modules:
            names = parsed.modules.split(',')
            print(ansi.success('RUNNING ') + 'tests in ' + str(names))
            suite = unittest.TestLoader().loadTestsFromNames(names)
        else:
            print(ansi.success('RUNNING ') + 'all tests')
            suite = unittest.defaultTestLoader.discover(env.tests_dir)

    result = unittest.TextTestRunner().run(suite)
    if not result.wasSuccessful():
        sys.exit(1)
    else:
        sys.exit(0)
    

def notebook(parsed, unknown):
    install_jupyter_kernel()
    args = [env.bin_jupyter, 'notebook'] + unknown
    print(ansi.success('JUPYTER') + ' ' + str(env.bin_jupyter))
    os.execv(env.bin_jupyter, args)


def install_darwin():
    install_gcc_5()
    install_pyenv()


def install_linux():
    install_pyenv()


def install_homebrew():
    if which('brew'):
        return
        
    print(ansi.success('INSTALL') + ' homebrew')
    subprocess.check_call((
        '/usr/bin/ruby',
        '-e',
        '"$(curl -fsSL https://raw.githubusercontent.com/Homebrew/install/master/install)"'
    ))


def install_pyenv():
    home = os.environ.get('HOME')
    if not home:
        return

    pyenv = os.path.join(home, '.pyenv')
    bin_pyenv = os.path.join(pyenv, 'bin', 'pyenv')
    virtualenv = os.path.join(pyenv, 'plugins', 'pyenv-virtualenv')
    if os.path.exists(bin_pyenv) and os.path.exists(virtualenv):
        return

    if os.path.exists(pyenv) and not os.path.isfile(bin_pyenv):
        print(ansi.warning() + ' pyenv executable is not present at %s' % bin_pyenv)
        while True:
            answer = input('Would you like to blow away ~/.pyenv and rebuild from scratch? [Y/n] ')
            if answer in ['', 'y', 'Y']:
                shutil.rmtree(pyenv)
                break
            elif answer in ['n', 'N']:
                sys.exit(ansi.error() + ' please fix pyenv before continuing')
            else:
                print('please enter Y or N')
                
    if not os.path.exists(pyenv):
        print(ansi.success('INSTALLING') + ' pyenv')
        subprocess.check_call((
            'git',
            'clone',
            'https://github.com/pyenv/pyenv.git',
            pyenv
        ))
    else:
        print(ansi.success('CHECK') + ' existing pyenv installation')
    env.pyenv = pyenv
    env.bin_pyenv = bin_pyenv
    env.set_python_version(env.python_version)

    if not os.path.exists(virtualenv):
        print(ansi.success('INSTALLING') + ' pyenv virtualenv')
        subprocess.check_call((
            'git',
            'clone',
            'https://github.com/pyenv/pyenv-virtualenv.git',
            virtualenv
        ))
    else:
        print(ansi.success('CHECK') + ' existing virtualenv installation')
    

def install_xcode():
    result = subprocess.call(('xcode-select', '--install'), stderr=subprocess.PIPE)
    if result > 0:
        print(ansi.success('CHECK') + ' xcode-select --install')
    else:
        print(ansi.success('INSTALL') + ' xcode command line tools')


def install_gcc_5():
    if which('gcc-5'):
        return

    install_homebrew()
    print(ansi.success('INSTALL') + ' gcc 5 for xgboost')
    subprocess.check_call(('brew', 'install', 'gcc@5'))


def install_bazel():
    if which('bazel'):
        return

    install_homebrew()
    print(ansi.success('INSTALL') + ' bazel for tensorflow')
    subprocess.check_call(('brew', 'install', 'bazel'))


def install_tensorflow():
    description = subprocess.check_output(
        (env.bin_python, '-m', 'pip', 'show', 'tensorflow')
    ).decode('utf-8')
    version = re.match(
        '.*^Version: ([^\n]+)', description, re.S | re.M
    ).group(1)
    if not version:
        sys.exit(ansi.error() + ' tensorflow is not in requirements.txt')
    
    print(ansi.success('NATIVE') + ' tensorflow ' + version)

    python_version = ''.join(env.python_version.split('.')[0:2])
    cached = os.path.join(
        env.pyenv,
        'cache',
        'tensorflow_pkg',
        'tensorflow-' + version + '-cp' + python_version + '*'
    )
    
    paths = glob.glob(cached)
    
    if not paths:
        build_tensorflow(version)
        paths = glob.glob(cached)

    path = paths[0]
    
    subprocess.check_call((env.bin_python, '-m', 'pip', 'uninstall', '-y', 'tensorflow'))
    print(ansi.success('INSTALL') + ' tensorflow native build')
    subprocess.check_call((env.bin_python, '-m', 'pip', 'install', path))


def build_tensorflow(version):
    install_bazel()
    print(ansi.success('BUILD') + ' tensorflow for this architecture')
    
    tensorflow_repo = os.path.join(env.pyenv, 'cache', 'tensorflow')
    cache = os.path.join(env.pyenv, 'cache', 'tensorflow_pkg')
    if not os.path.exists(tensorflow_repo):
        subprocess.check_call((
            'git',
            'clone',
            'https://github.com/tensorflow/tensorflow',
            tensorflow_repo
        ))
    
    subprocess.check_call(
        ('git', 'checkout', '--', '.'),
        cwd=tensorflow_repo
    )
    subprocess.check_call(
        ('git', 'checkout', 'master'),
        cwd=tensorflow_repo
    )
    subprocess.check_call(
        ('git', 'pull'),
        cwd=tensorflow_repo
    )
    subprocess.check_call(
        ('git', 'checkout', 'v' + version),
        cwd=tensorflow_repo
    )
    major, minor, patch = env.python_version.split('.')
    lib = os.path.join('lib', 'python' + major + '.' + minor, 'site-packages')
    new_env = {
        'PATH': os.environ['PATH'],
        'PYTHON_BIN_PATH': env.bin_python,
        'PYTHON_LIB_PATH': os.path.join(env.prefix, lib),
        'TF_NEED_MKL': '0',
        'CC_OPT_FLAGS': '-march=native -O2',
        'TF_NEED_JEMALLOC': '1',  # only available on linux regardless
        'TF_NEED_GCP': '0',
        'TF_NEED_HDFS': '0',
        'TF_ENABLE_XLA': '0',
        'TF_NEED_VERBS': '0',
        'TF_NEED_OPENCL': '0',
        'TF_NEED_S3': '0',
        'TF_NEED_GDR': '0',
        'TF_NEED_CUDA': '0',  # TODO enable CUDA when appropriate
        'TF_CUDA_CLANG': '1',
        'TF_CUDA_VERSION': '8.0.61',
        'CUDA_TOOLKIT_PATH': '/usr/local/cuda',
        'CUDNN_INSTALL_PATH': '/usr/local/cuda',
        'TF_CUDNN_VERSION': '5.1.10',
        'TF_CUDA_CLANG': '/usr/bin/gcc',
        'TF_CUDA_COMPUTE_CAPABILITIES': '3.5,5.2',
        'TF_NEED_MPI': '0'
    }
    subprocess.check_call(('./configure',), cwd=tensorflow_repo, env=new_env)
    # TODO remove this hack when tensorflow fixes their build
    # https://github.com/tensorflow/tensorflow/issues/12979
    pip = subprocess.Popen(
        (
            'sed',
            '-i',
            "'\@https://github.com/google/protobuf/archive/0b059a3d8a8f8aa40dde7bea55edca4ec5dfea66.tar.gz@d'",
            'tensorflow/workspace.bzl'
        ),
        stdin=None,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    (stdout, stderr) = pip.communicate()
    pip.wait()

    subprocess.check_call((
        'bazel',
        'build',
        '--config=opt',
        # '--config=cuda',  TODO enable CUDA when appropriate
        'tensorflow/tools/pip_package:build_pip_package',
    ), cwd=tensorflow_repo)
    
    subprocess.check_call((
        'bazel-bin/tensorflow/tools/pip_package/build_pip_package',
        cache
    ), cwd=tensorflow_repo)


def install_python_version():
    if env.launched():
        return
    
    if not env.python_version:
        env.set_python_version('.'.join(sys.version_info))
        print(ansi.warning() + ' %s does not exist. Creating with %s' %
              (env.version_path, env.python_version))
        with open(env.version_path, 'w', encoding='utf-8') as f:
            f.write(env.python_version + os.linesep)
    
    if not env.pyenv:
        sys.exit(
            ansi.error() + ' pyenv is not installed. Lore is broken. try:\n'
            ' $ pip uninstall lore && pip install lore\n'
        )
    
    versions = subprocess.check_output(
        (env.bin_pyenv, 'versions', '--bare')
    ).decode('utf-8').split(os.linesep)
    if env.python_version not in versions:
        print(ansi.success('INSTALL') + ' python %s' % env.python_version)
        if platform.system() == 'Darwin':
            install_xcode()
        subprocess.check_call(('git', '-C', env.pyenv, 'pull'))
        subprocess.check_call((env.bin_pyenv, 'install', env.python_version))
        subprocess.check_call((env.bin_pyenv, 'rehash'))


def create_virtual_env():
    if env.pyenv:
        try:
            os.unlink(os.path.join(env.pyenv, 'versions', env.project))
        except OSError as e:
            pass
    
    if os.path.exists(env.bin_python):
        return

    print(ansi.success('CREATE') + ' virtualenv: %s' % env.project)
    subprocess.check_call((
        env.bin_pyenv,
        'virtualenv',
        env.python_version,
        env.project
    ))
    subprocess.check_call([env.bin_python, '-m', 'pip', 'install', 'six'])


def install_requirements(args):
    source = env.requirements
    if not os.path.exists(source):
        sys.exit(
            ansi.error() + ' %s is missing. You should check it in to version '
                             'control.' % ansi.underline(source)
        )
    
    pip_install(source, args)
    freeze_requirements()
    install_jupyter_kernel()
    
    
def install_jupyter_kernel():
    import jupyter_core.paths
    if os.path.exists(os.path.join(jupyter_core.paths.jupyter_data_dir(), 'kernels', env.project)):
        return

    print(ansi.success('INSTALL') + ' jupyter kernel')
    subprocess.check_call((
        env.bin_python,
        '-m',
        'ipykernel',
        'install',
        '--user',
        '--name=' + env.project
    ))


def freeze_requirements():
    source = env.requirements
    
    print(ansi.success('EXECUTE') + ' ' + env.bin_python + ' -m pip freeze -r ' + source)
    vcs = split_vcs_lines()
    pip = subprocess.Popen(
        (env.bin_python, '-m', 'pip', 'freeze', '-r', source),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    (stdout, stderr) = pip.communicate()
    pip.wait()
    
    restore_vcs_lines(vcs)

    present = stdout.decode('utf-8').split(os.linesep)
    errors = stderr.decode('utf-8').split(os.linesep)
    missing = [line for line in errors if 'package is not installed' in line]
    regex = re.compile(r'contains ([\w\-\_]+)')
    needed = [m.group(1).lower() for l in missing for m in [regex.search(l)] if m]

    added_index = present.index('## The following requirements were added by pip freeze:')
    if added_index:
        added = present[added_index + 1:-1]
        present = set(present[0:added_index])
        safe = set()
        unsafe = set()
        for package in added:
            name = package.split('==')[0]

            for bad in vcs:
                if name in bad:
                    unsafe.add(package)
                    continue
            
            if name.lower() in needed:
                needed.remove(name.lower())
            
            safe.add(package)
        present |= safe
        present -= unsafe
    
    if needed:
        args = [env.bin_python, '-m', 'pip', 'install'] + needed
        print(ansi.success('EXECUTE ') + ' '.join(args))
        subprocess.check_call(args)
        return freeze_requirements()

    if unsafe:
        if vcs:
            print(
                ansi.warning() + ' Non pypi packages were detected in your ' +
                ansi.underline('requirements.txt') + ' that can not be '
                'completely frozen by pip. ' + os.linesep + os.linesep +
                os.linesep.join(vcs)
            )
    
        print(
            ansi.info() + ' You should check the following packages in to ' +
            ansi.underline('requirements.txt') + ' or `lore pip uninstall` them'
        )
        if vcs:
            print(ansi.warning('unless') + ' they are covered by the previously listed packages'
            ' that pip can not freeze.')
        print(
            os.linesep + os.linesep.join(unsafe) + os.linesep
        )

    with open(source, 'w', encoding='utf-8') as f:
        f.write(os.linesep.join(sorted(present, key=lambda s: s.lower())).strip() + os.linesep)
        if vcs:
            f.write(os.linesep.join(vcs) + os.linesep)


def split_vcs_lines():
    with open(env.requirements, 'r', encoding='utf-8') as f:
        lines = f.readlines()
    
    vcs = [line for line in lines if
           re.match(r'^(-e )?(git|svn|hg|bzr).*', line)]
    if not vcs:
        return vcs
    
    if os.path.exists(env.requirements_vcs):
        with open(env.requirements_vcs, 'r', encoding='utf-8') as f:
            new = set(f.readlines())
            vcs = list(set(vcs).union(new))
    
    lines = list(set(lines) - set(vcs))
    with open(env.requirements, 'w', encoding='utf-8') as f:
        f.write(''.join(sorted(lines)))
    
    with open(env.requirements_vcs, 'w', encoding='utf-8') as f:
        f.write(''.join(sorted(vcs)))
    return vcs


def restore_vcs_lines(vcs):
    if not os.path.exists(env.requirements_vcs):
        return
    with open(env.requirements, 'r', encoding='utf-8') as f:
        original = f.read()
    with open(env.requirements, 'w', encoding='utf-8') as f:
        f.write(''.join(vcs) + original)
    os.remove(env.requirements_vcs)


def pip_install(path, args):
    if not os.path.exists(path):
        return
    
    pip_args = [env.bin_python, '-m', 'pip', 'install', '-r', path]
    if hasattr(args, 'upgrade') and args.upgrade:
        pip_args += ['--upgrade', '--upgrade-strategy=eager']
    print(ansi.success('EXECUTE ') + ' '.join(pip_args))
    try:
        subprocess.check_call(pip_args)
    except subprocess.CalledProcessError:
        sys.exit(
            ansi.error() + ' could not:\n $ lore pip install -r %s\nPlease try '
                           'installing failed packages manually, or upgrade failed '
                           'packages with:\n $ lore install --upgrade ' % path
        )


if __name__ == '__main__':
    try:
        main()
    except:
        e = sys.exc_info()
        for i in e:
            print(i)
        import traceback
        traceback.print_tb(e[2])



import os
import lore
from lore.stores.disk import Disk
from lore.stores.ram import Ram


cache = Ram()
query_cache = Disk(os.path.join(lore.env.data_dir, 'query_cache'))


def cached(func):
    global cache
    return _cached(func, cache)


def query_cached(func):
    global query_cache
    return _cached(func, query_cache)


def _cached(func, store):
    def wrapper(self, *args, **kwargs):
        cache = kwargs.pop('cache', False)
        if not cache:
            return func(self, *args, **kwargs)
        
        key = store.key(instance=self, caller=func, *args, **kwargs)
        if key not in store:
            store[key] = func(self, *args, **kwargs)
        return store[key]
    
    return wrapper



import os
from os.path import join, dirname
import h5py

import lore
from lore.util import timer


class Base(lore.models.base.Base):
    def __init__(self, pipeline, estimator):
        super(Base, self).__init__(pipeline, estimator)
    
    def weights_path(self):
        return join(self.fitting_path(), 'weights.h5')
    
    def checkpoint_path(self):
        return join(self.fitting_path(), 'checkpoints/{epoch}.h5')
    
    def tensorboard_path(self):
        return join(self.fitting_path(), 'tensorboard')

    def timeline_path(self):
        return join(self.fitting_path(), 'timeline.json')

    def remote_weights_path(self):
        return join(self.remote_path(), 'weights.h5')

    @property
    def fitting(self):
        return self._fitting
    
    @fitting.setter
    def fitting(self, value):
        self._fitting = value
        if self._fitting is not None:
            if not os.path.exists(dirname(self.checkpoint_path())):
                os.makedirs(dirname(self.checkpoint_path()))
                
            if not os.path.exists(dirname(self.tensorboard_path())):
                os.makedirs(dirname(self.tensorboard_path()))
    
    def save(self, stats=None):
        super(Base, self).save(stats)
        
        with timer('save weights:'):
            # Only save weights, because saving named layers that have shared
            # weights causes an error on reload
            self.estimator.keras.save_weights(self.weights_path())
        
        # Patch for keras 2 models saved with optimizer weights:
        # https://github.com/gagnonlg/explore-ml/commit/c05b01076c7eb99dae6a480a05ac14765ef08e4b
        with h5py.File(self.weights_path(), 'a') as f:
            if 'optimizer_weights' in f.keys():
                del f['optimizer_weights']
    
    @classmethod
    def load(cls, fitting=None):
        model = super(Base, cls).load(fitting)
        
        if hasattr(model, 'estimator'):
            # HACK to set estimator model, and model serializer
            model.estimator = model.estimator
            
            # Rely on build + load_weights rather than loading the named layers
            # w/ Keras for efficiency (and also because it causes a
            # deserialization issue) as of Keras 2.0.4:
            # https://github.com/fchollet/keras/issues/5442
            model.estimator.build()
            
            with timer('load weights:'):
                model.estimator.keras.load_weights(model.weights_path())
        else:
            model.build()
            with timer('load weights:'):
                model.keras.load_weights(model.weights_path())
        
        return model
    
    def upload(self):
        super(Base, self).upload()
        lore.io.upload(self.weights_path(), self.remote_weights_path())
    
    @classmethod
    def download(cls, fitting=0):
        model = cls(None, None)
        model.fitting = fitting
        lore.io.download(model.weights_path(), model.remote_weights_path())
        return super(Base, cls).download(fitting)



from lore.stores.base import Base
import lore.io


class Redis(Base):
    def __init__(self, redis_conn):
        self.r = redis_conn

    def __getitem__(self, key):
        return self.r.get(key)

    def __setitem__(self, key, value):
        self.r.set(key, value)

    def __delitem__(self, key):
        self.r.delete(key)

    def __contains__(self, key):
        return self.r.get(key) != None

    def __len__(self):
        raise "Operation can be expensive. Aborting"

    def keys(self):
        raise "Operation can be expensive. Aborting"

    def values(self):
        raise "Operation can be expensive. Aborting"

    def batch_get(self, keys):
        return self.r.mget(keys)

    def batch_set(self, data_dict):
        self.r.mset(data_dict)




import os
import pickle

from lore.stores.base import Base
from lore.util import timer


class Disk(Base):
    EXTENSION = '.pickle'
    
    def __init__(self, dir):
        self.dir = dir
        self.limit = None
        if not os.path.exists(self.dir):
            os.makedirs(self.dir)
    
    def __getitem__(self, key):
        if key in self:
            with timer('read %s:' % key):
                with open(self._path(key), 'rb') as f:
                    return pickle.load(f)
        return None
    
    def __setitem__(self, key, value):
        with timer('write %s:' % key):
            with open(self._path(key), 'wb') as f:
                pickle.dump(value, f, pickle.HIGHEST_PROTOCOL)

        if self.limit is not None:
            if os.path.getsize(self._path(key)) > self.limit:
                raise MemoryError('disk cache limit exceeded by single key: %s' % key)
    
            with timer('evict: %s' % key):
                while self.size() > self.limit:
                    del self[self.lru()]
    
    def __delitem__(self, key):
        os.remove(self._path(key))
    
    def __contains__(self, key):
        return os.path.isfile(self._path(key))
    
    def __len__(self):
        return len(self.keys())

    def batch_get(self, keys):
        result = {}
        for key in keys:
            result[key] = self.__getitem__(key)
        return result

    def batch_set(self, data_dict):
        for key, value in data_dict.items():
            self.__setitem__(key, value)

    def size(self):
        return sum(os.path.getsize(f) for f in self.values())
    
    def keys(self):
        return [self._key(f) for f in os.listdir(self.dir)]
    
    def values(self):
        return [os.path.join(self.dir, f) for f in os.listdir(self.dir)]
    
    def lru(self):
        files = sorted(self.values(), key=lambda f: os.stat(f).st_atime)
        
        if not files:
            return None
        
        return self._key(files[0])
    
    def _path(self, key):
        return os.path.join(self.dir, key + self.EXTENSION)
    
    def _key(self, path):
        return os.path.basename(path)[0:-len(self.EXTENSION)]



from __future__ import absolute_import
import atexit
import logging

import pandas
import keras
import keras.backend
from keras.callbacks import EarlyStopping, TensorBoard, TerminateOnNaN
from keras.layers import Input, Embedding, Dense, Reshape, Concatenate, Dropout, SimpleRNN, Flatten
from keras.optimizers import Adam
from sklearn.base import BaseEstimator
import tensorflow
from tensorflow.python.client.timeline import Timeline

import lore.io
from lore.callbacks import ReloadBest
from lore.encoders import Continuous
from lore.pipelines import Observations
from lore.util import timed

from tensorflow.python.client import device_lib
gpus = [x.name for x in device_lib.list_local_devices() if x.device_type == 'GPU']
if len(gpus) > 0:
    from keras.layers import CuDNNLSTM as LSTM
    from keras.layers import CuDNNGRU as GRU
else:
    from keras.layers import LSTM
    from keras.layers import GRU

logger = logging.getLogger(__name__)


# prevents random gc exception at exit
atexit.register(keras.backend.clear_session)


class Keras(BaseEstimator):
    def __init__(
            self,
            model=None,
            embed_size=10,
            sequence_embedding='flatten',
            sequence_embed_size=10,
            hidden_width=1024,
            hidden_layers=4,
            layer_shrink=0.5,
            dropout=0,
            batch_size=32,
            learning_rate=0.001,
            decay=0.,
            optimizer=None,
            hidden_activation='relu',
            hidden_activity_regularizer=None,
            hidden_bias_regularizer=None,
            hidden_kernel_regularizer=None,
            monitor='val_acc',
            loss='categorical_crossentropy',
            towers=1,
    ):
        super(Keras, self).__init__()
        self.towers = towers
        self.embed_size = embed_size
        self.sequence_embed_size = 10
        self.hidden_width = hidden_width
        self.hidden_layers = hidden_layers
        self.layer_shrink = layer_shrink
        self.dropout = dropout
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.decay = decay
        self.optimizer = optimizer
        self.hidden_activation = hidden_activation
        self.hidden_activity_regularizer = hidden_activity_regularizer
        self.hidden_bias_regularizer = hidden_bias_regularizer
        self.hidden_kernel_regularizer = hidden_kernel_regularizer
        self.monitor = monitor
        self.loss = loss
        self.keras = None
        self.history = None
        self.session = None
        self.model = model
        self.sequence_embedding = sequence_embedding
        self.sequence_embed_size = sequence_embed_size
        
    
    def __getstate__(self):
        state = super(Keras, self).__getstate__()
        # bloat can be restored via self.__init__() + self.build()
        for bloat in [
            'keras',
            'optimizer',
            'session',
            'model'
        ]:
            state[bloat] = None
        return state
    
    @property
    def description(self):
        return '\n'.join([
            '\n  %s' % self.__module__ + '.' + self.__class__.__name__,
            '==========================================================',
            '| embed | hidden | layer | layer  |         | model      |',
            '| size  | layers | width | shrink | dropout | parameters |',
            '----------------------------------------------------------',
            '| %5i | %6i | %5i | %6.4f | %7.5f | %10i |' % (
                self.embed_size,
                self.hidden_layers,
                self.hidden_width,
                self.layer_shrink,
                self.dropout,
                self.keras.count_params()
            ),
            '=========================================================='
        ])
    
    def callbacks(self):
        return []

    @timed(logging.INFO)
    def build(self, log_device_placement=False):
        keras.backend.clear_session()
        self.session = tensorflow.Session(config=tensorflow.ConfigProto(log_device_placement=log_device_placement))
        keras.backend.set_session(self.session)
        with self.session.as_default():
            inputs = self.build_inputs()
            outputs = []
            for i in range(self.towers):
                if len(gpus) > 0:
                    chip = 'gpu'
                    node = i % len(gpus)
                else:
                    chip = 'cpu'
                    node = 0
                with tensorflow.device('/%s:%i' % (chip, node)):
                    embedding_layer = self.build_embedding_layer(inputs, i)
                    hidden_layers = self.build_hidden_layers(embedding_layer, i)
                    outputs.append(self.build_output_layer(hidden_layers, i))
        
            self.keras = keras.models.Model(inputs=list(inputs.values()), outputs=outputs)
            self.optimizer = Adam(lr=self.learning_rate, decay=self.decay)
            self.keras._make_predict_function()
        logger.info('\n\n' + self.description + '\n\n')

    @timed(logging.INFO)
    def build_inputs(self):
        inputs = {}
        for encoder in self.model.pipeline.encoders:
            if hasattr(encoder, 'sequence_length'):
                for i in range(encoder.sequence_length):
                    inputs[encoder.sequence_name(i)] = Input(shape=(1,), name=encoder.sequence_name(i))
            else:
                inputs[encoder.name] = Input(shape=(1,), name=encoder.name)
        return inputs

    @timed(logging.INFO)
    def build_embedding_layer(self, inputs, tower):
        embeddings = {}
        for encoder in self.model.pipeline.encoders:
            embed_name = str(tower) + '_embed_' + encoder.name
            embed_size = encoder.embed_scale * self.embed_size

            if isinstance(encoder, Continuous):
                embedding = Dense(embed_size, activation='relu', name=embed_name)
            else:
                embedding = Embedding(encoder.cardinality(), embed_size, name=embed_name)

            if hasattr(encoder, 'sequence_length'):
                embeddings[embed_name] = self.build_sequence_embedding(encoder, embedding, inputs, embed_name)
            else:
                embeddings[embed_name] = Reshape(target_shape=(embed_size,))(embedding(inputs[encoder.name]))

        return Concatenate()(list(embeddings.values()))
    
    def build_sequence_embedding(self, encoder, embedding, inputs, embed_name, suffix=''):
        embed_size = encoder.embed_scale * self.embed_size
    
        sequence = []
        for i in range(encoder.sequence_length):
            sequence.append(embedding(inputs[encoder.sequence_name(i, suffix)]))
        embed_sequence = Concatenate(name=embed_name + '_sequence' + suffix)(sequence)
    
        if self.sequence_embedding == 'flatten':
            embedding = Flatten(name=embed_name + '_flatten' + suffix)(embed_sequence)
        else:
            sequence_embed_size = encoder.embed_scale * self.sequence_embed_size
            shaped_sequence = Reshape(target_shape=(encoder.sequence_length, embed_size))(embed_sequence)
            if self.sequence_embedding == 'lstm':
                embedding = LSTM(sequence_embed_size, name=embed_name + '_lstm' + suffix)(shaped_sequence)
            elif self.sequence_embedding == 'gru':
                embedding = GRU(sequence_embed_size, name=embed_name + '_gru' + suffix)(shaped_sequence)
            elif self.sequence_embedding == 'simple_rnn':
                embedding = SimpleRNN(sequence_embed_size, name=embed_name + '_rnn' + suffix)(shaped_sequence)
            else:
                raise ValueError("Unknown sequence_embedding type: %s" % self.sequence_embedding)
        return embedding

    @timed(logging.INFO)
    def build_hidden_layers(self, input_layer, tower):
        hidden_layers = input_layer

        hidden_width = self.hidden_width
        for i in range(self.hidden_layers):
            hidden_layers = Dense(int(hidden_width),
                                  activation=self.hidden_activation,
                                  activity_regularizer=self.hidden_activity_regularizer,
                                  kernel_regularizer=self.hidden_kernel_regularizer,
                                  bias_regularizer=self.hidden_bias_regularizer,
                                  name='%i_hidden_%i' % (tower, i))(hidden_layers)
            if self.dropout > 0:
                hidden_layers = Dropout(self.dropout)(hidden_layers)
            if self.layer_shrink < 1:
                hidden_width *= self.layer_shrink
            else:
                hidden_width -= self.layer_shrink
            hidden_width = max(1, hidden_width)
            
        return hidden_layers
    
    @timed(logging.INFO)
    def build_output_layer(self, hidden_layers, tower):
        return Dense(1, activation='sigmoid', name='%i_output' % tower)(hidden_layers)
    
    @timed(logging.INFO)
    def fit(self, x, y, validation_data=None, epochs=100, patience=0, verbose=None, min_delta=0, tensorboard=False, timeline=False):

        if validation_data is None:
            validation_data = self.model.pipeline.encoded_validation_data

        if isinstance(x, pandas.DataFrame):
            x = x.to_dict(orient='series')
        
        if isinstance(validation_data.x, pandas.DataFrame):
            validation_data = Observations(
                x=validation_data.x.to_dict(orient='series'),
                y=validation_data.y
            )
            
        if not self.keras or not self.optimizer:
            self.build()
            
        with self.session.as_default():
            if timeline:
                run_metadata = tensorflow.RunMetadata()
            else:
                run_metadata = None
            self.keras.compile(
                loss=self.loss,
                optimizer=self.optimizer,
                options=tensorflow.RunOptions(trace_level=tensorflow.RunOptions.FULL_TRACE),
                run_metadata=run_metadata
            )
        if verbose is None:
            verbose = 1 if lore.env.name == lore.env.DEVELOPMENT else 0
        
        logger.info(
            '\n'.join([
                '\n\n\n  Fitting',
                '==============================',
                '| batch | learning |         |',
                '| size  | rate     |   decay |',
                '------------------------------',
                '| %5i | %8.6f | %7.5f |' % (
                    self.batch_size,
                    self.learning_rate,
                    self.decay,
                ),
                '==============================\n\n'
            ])
        )
        
        reload_best = ReloadBest(
            filepath=self.model.checkpoint_path(),
            monitor=self.monitor,
            mode='auto',
        )
        
        callbacks = self.callbacks()
        callbacks += [
            reload_best,
            TerminateOnNaN(),
            EarlyStopping(
                monitor=self.monitor,
                min_delta=min_delta,
                patience=patience,
                verbose=verbose,
                mode='auto',
            ),
        ]
        if tensorboard:
            callbacks += [TensorBoard(
                log_dir=self.model.serializer.tensorboard_path,
                histogram_freq=1,
                batch_size=self.batch_size,
                write_graph=True,
                write_grads=True,
                write_images=True,
                embeddings_freq=1,
                embeddings_metadata=None
            )]
        
        with self.session.as_default():
            self.history = self.keras.fit(
                x=x,
                y=[y] * self.towers,
                validation_data=Observations(x=validation_data.x, y=[validation_data.y] * self.towers),
                batch_size=self.batch_size,
                epochs=epochs,
                verbose=verbose,
                callbacks=callbacks
            ).history

        if timeline:
            with open(self.model.timeline_path(), 'w') as f:
                f.write(Timeline(step_stats=run_metadata.step_stats).generate_chrome_trace_format())

        return {
            'epochs': len(self.history),
            'train': reload_best.train_loss,
            'validate': reload_best.validate_loss,
            'timeline': timeline,
        }

    @timed(logging.DEBUG)
    def predict(self, dataframe):
        if isinstance(dataframe, pandas.DataFrame):
            dataframe = dataframe.to_dict(orient='series')
        with self.session.as_default():
            return self.keras.predict(dataframe, batch_size=self.batch_size)
    
    @timed(logging.INFO)
    def score(self, x, y):
        if isinstance(x, pandas.DataFrame):
            x = x.to_dict(orient='series')
        with self.session.as_default():
            return 1 / self.keras.evaluate(x, y, batch_size=self.batch_size)



"""
Lore Environment

Key attributes and paths for this project
"""
from __future__ import absolute_import

try:
    import configparser
except ImportError:
    configparser = None
    
import glob
import os
import re
import sys
from io import open
import pkg_resources
from pkg_resources import DistributionNotFound, VersionConflict
import socket

from lore import ansi

TEST = 'test'
DEVELOPMENT = 'development'
PRODUCTION = 'production'


def read_version(path):
    version = None
    if os.path.exists(path):
        with open(path, 'r', encoding='utf-8') as f:
            version = f.read().strip()
    
    if version:
        return re.sub(r'^python-', '', version)
    
    return version


version_path = 'runtime.txt'
python_version = None
python_version_info = None
if 'LORE_ROOT' in os.environ:
    root = os.environ.get('LORE_ROOT')
else:
    root = os.getcwd()
    while True:
        python_version = read_version(os.path.join(root, 'runtime.txt'))
        if python_version:
            break
        
        python_version = read_version(os.path.join(root, '.python-version'))
        if python_version:
            break
        
        root = os.path.dirname(root)
        if root == '/':
            root = os.getcwd()
            break

lib = os.path.join(root, 'lib')
if lib not in sys.path:
    sys.path.insert(0, lib)

# Load environment variables from disk
for var in glob.glob('/conf/env/*'):
    if os.path.isfile(var):
        os.environ[os.path.basename(var)] = open(var, encoding='utf-8').read()

env_file = os.path.join(root, '.env')
if os.path.isfile(env_file):
    from dotenv import load_dotenv
    load_dotenv(env_file)

if len(sys.argv) > 1 and sys.argv[1] == 'test':
    default_name = TEST
else:
    default_name = DEVELOPMENT

host = socket.gethostname()
home = os.environ.get('HOME', root)
name = os.environ.get('LORE_ENV', default_name)
project = os.environ.get('LORE_PROJECT', root.split(os.sep)[-1])
sys.path = [root] + sys.path
work_dir = 'tests' if name == TEST else os.environ.get('WORK_DIR', root)
models_dir = os.path.join(work_dir, 'models')
data_dir = os.path.join(work_dir, 'data')
log_dir = os.path.join(root if name == TEST else work_dir, 'logs')
tests_dir = os.path.join(root, 'tests')
color = {
    DEVELOPMENT: ansi.GREEN,
    TEST: ansi.BLUE,
    PRODUCTION: ansi.RED,
}.get(name, ansi.YELLOW)


pyenv = os.path.join(home, '.pyenv')
if os.path.exists(pyenv):
    pyenv = os.path.realpath(pyenv)
    bin_pyenv = os.path.join(pyenv, 'bin', 'pyenv')
else:
    pyenv = None
    bin_pyenv = None

prefix = None
bin_python = None
bin_lore = None
bin_jupyter = None
requirements = os.path.join(root, 'requirements.txt')
requirements_vcs = os.path.join(root, 'requirements.vcs.txt')


def set_python_version(version):
    """Set the python version for this lore project, to establish the location
    of key binaries.
    
    :param version:
    :type version: unicode
    """
    global python_version
    global python_version_info
    global prefix
    global bin_python
    global bin_lore
    global bin_jupyter
    
    python_version = version
    if python_version:
        python_version_info = tuple([int(i) if i.isdigit() else i for i in version.split('.')])
        if pyenv:
            prefix = os.path.join(
                pyenv,
                'versions',
                python_version,
                'envs',
                project
            )
        else:
            prefix = os.path.realpath(sys.prefix)
            
        bin_python = os.path.join(prefix, 'bin', 'python' + '.'.join([str(i) for i in python_version_info[0:2]]))
        if not os.path.exists(bin_python):
            bin_python = os.path.join(prefix, 'bin', 'python' + str(python_version_info[0]))
        if not os.path.exists(bin_python):
            bin_python = os.path.join(prefix, 'bin', 'python')
        bin_lore = os.path.join(prefix, 'bin', 'lore')
        bin_jupyter = os.path.join(prefix, 'bin', 'jupyter')
    else:
        python_version_info = []
        prefix = None
        bin_python = None
        bin_lore = None
        bin_jupyter = None

set_python_version(python_version)


def exists():
    """Test whether the current working directory has a valid lore environment.
    
    :return:  bool True if the environment is valid
    """
    return python_version is not None


def launched():
    """Test whether the current python environment is the correct lore env.

    :return:  bool True if the environment is launched
    """
    if not prefix:
        return False
    
    return os.path.realpath(sys.prefix) == os.path.realpath(prefix)


def validate():
    if not os.path.exists(os.path.join(root, project, '__init__.py')):
        sys.exit(ansi.error() + ' Python module not found. Do you need to change $LORE_PROJECT from "%s"?' % project)

    if exists():
        return

    if len(sys.argv) > 1:
        command = sys.argv[1]
    else:
        command = 'lore'
    sys.exit(
        ansi.error() + ' %s is only available in lore '
                       'project directories (missing %s)' % (
            ansi.bold(command),
            ansi.underline(version_path)
        )
    )


def launch():
    if launched():
        check_version()
        os.chdir(root)
        return
    
    if not os.path.exists(bin_lore):
        missing = ' %s virtualenv is missing.' % project
        if '--launched' in sys.argv:
            sys.exit(ansi.error() + missing + ' Please check for errors during:\n $ lore install\n')
        else:
            print(ansi.warning() + missing)
            import lore.__main__
            lore.__main__.install(None, None)
        
    reboot('--env-launched')
    

def reboot(*args):
    args = list(sys.argv) + list(args)
    if args[0] == 'python' or not args[0]:
        args[0] = bin_python
    elif args[0][-4:] == 'lore':
        args[0] = bin_lore
    os.execv(args[0], args)


def check_version():
    if sys.version_info[0:3] == python_version_info[0:3]:
        return

    sys.exit(
        ansi.error() + ' your virtual env points to the wrong python version. '
                       'This is likely because you used a python installer that clobbered '
                       'the system installation, which breaks virtualenv creation. '
                       'To fix, check this symlink, and delete the installation of python '
                       'that it is brokenly pointing to, then delete the virtual env itself '
                       'and rerun lore install: ' + os.linesep + os.linesep + bin_python +
        os.linesep
    )


def check_requirements():
    if not os.path.exists(requirements):
        sys.exit(
            ansi.error() + ' %s is missing. Please check it in.' % ansi.underline(requirements)
        )
    
    with open(requirements, 'r', encoding='utf-8') as f:
        dependencies = f.readlines()
    
    vcs = [d for d in dependencies if re.match(r'^(-e )?(git|svn|hg|bzr).*', d)]
    
    dependencies = list(set(dependencies) - set(vcs))
    
    missing = []
    try:
        pkg_resources.require(dependencies)
    except (pkg_resources.ContextualVersionConflict, DistributionNotFound, VersionConflict) as error:
        missing.append(str(error))
    except pkg_resources.RequirementParseError:
        pass
    
    if missing:
        missing = ' missing requirement:\n  ' + os.linesep.join(missing)
        if '--env-checked' in sys.argv:
            sys.exit(ansi.error() + missing + '\nRequirement installation failure, please check for errors in:\n $ lore install\n')
        else:
            print(ansi.warning() + missing)
            import lore.__main__
            lore.__main__.install_requirements(None)
            reboot('--env-checked')
    else:
        return True


def get_config(path):
    if configparser is None:
        return None
    
    # Check for env specific configs first
    if os.path.exists(os.path.join(root, 'config', name, path)):
        path = os.path.join(root, 'config', name, path)
    else:
        path = os.path.join(root, 'config', path)

    if not os.path.isfile(path):
        return None
    
    conf = open(path, 'rt').read()
    conf = os.path.expandvars(conf)
    
    config = configparser.SafeConfigParser()
    if sys.version_info[0] == 2:
        from io import StringIO
        config.readfp(StringIO(unicode(conf)))
    else:
        config.read_string(conf)
    return config

aws_config = get_config('aws.cfg')
database_config = get_config('database.cfg')
redis_config = get_config('redis.cfg')



#The MIT License (MIT)
#
#Copyright (c) 2015 Andre Queiroz
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in
#all copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#THE SOFTWARE.
#
# Holt-Winters algorithms to forecasting
# Coded in Python 2 by: Andre Queiroz
# Description: This module contains three exponential smoothing algorithms. They are Holt's linear trend method and Holt-Winters seasonal methods (additive and multiplicative).
# References:
#  Hyndman, R. J.; Athanasopoulos, G. (2013) Forecasting: principles and practice. http://otexts.com/fpp/. Accessed on 07/03/2013.
#  Byrd, R. H.; Lu, P.; Nocedal, J. A Limited Memory Algorithm for Bound Constrained Optimization, (1995), SIAM Journal on Scientific and Statistical Computing, 16, 5, pp. 1190-1208.

# https://gist.github.com/andrequeiroz/5888967

from __future__ import division
from sys import exit
from math import sqrt
from numpy import array
from scipy.optimize import fmin_l_bfgs_b

def RMSE(params, *args):

  Y = args[0]
  type = args[1]
  rmse = 0

  if type == 'linear':

    alpha, beta = params
    a = [Y[0]]
    b = [Y[1] - Y[0]]
    y = [a[0] + b[0]]

    for i in range(len(Y)):

      a.append(alpha * Y[i] + (1 - alpha) * (a[i] + b[i]))
      b.append(beta * (a[i + 1] - a[i]) + (1 - beta) * b[i])
      y.append(a[i + 1] + b[i + 1])

  else:

    alpha, beta, gamma = params
    m = args[2]   
    a = [sum(Y[0:m]) / float(m)]
    b = [(sum(Y[m:2 * m]) - sum(Y[0:m])) / m ** 2]

    if type == 'additive':

      s = [Y[i] - a[0] for i in range(m)]
      y = [a[0] + b[0] + s[0]]

      for i in range(len(Y)):

        a.append(alpha * (Y[i] - s[i]) + (1 - alpha) * (a[i] + b[i]))
        b.append(beta * (a[i + 1] - a[i]) + (1 - beta) * b[i])
        s.append(gamma * (Y[i] - a[i] - b[i]) + (1 - gamma) * s[i])
        y.append(a[i + 1] + b[i + 1] + s[i + 1])

    elif type == 'multiplicative':

      s = [Y[i] / a[0] for i in range(m)]
      y = [(a[0] + b[0]) * s[0]]

      for i in range(len(Y)):

        a.append(alpha * (Y[i] / s[i]) + (1 - alpha) * (a[i] + b[i]))
        b.append(beta * (a[i + 1] - a[i]) + (1 - beta) * b[i])
        s.append(gamma * (Y[i] / (a[i] + b[i])) + (1 - gamma) * s[i])
        y.append((a[i + 1] + b[i + 1]) * s[i + 1])

    else:

      exit('Type must be either linear, additive or multiplicative')
    
  rmse = sqrt(sum([(m - n) ** 2 for m, n in zip(Y, y[:-1])]) / len(Y))

  return rmse

def linear(x, fc, alpha = None, beta = None):

  Y = x[:]

  if (alpha == None or beta == None):

    initial_values = array([0.3, 0.1])
    boundaries = [(0, 1), (0, 1)]
    type = 'linear'

    parameters = fmin_l_bfgs_b(RMSE, x0 = initial_values, args = (Y, type), bounds = boundaries, approx_grad = True)
    alpha, beta = parameters[0]

  a = [Y[0]]
  b = [Y[1] - Y[0]]
  y = [a[0] + b[0]]
  rmse = 0

  for i in range(len(Y) + fc):

    if i == len(Y):
      Y.append(a[-1] + b[-1])

    a.append(alpha * Y[i] + (1 - alpha) * (a[i] + b[i]))
    b.append(beta * (a[i + 1] - a[i]) + (1 - beta) * b[i])
    y.append(a[i + 1] + b[i + 1])

  rmse = sqrt(sum([(m - n) ** 2 for m, n in zip(Y[:-fc], y[:-fc - 1])]) / len(Y[:-fc]))

  return Y[-fc:], alpha, beta, rmse

def additive(x, m, fc, alpha = None, beta = None, gamma = None):

  Y = x[:]

  if (alpha == None or beta == None or gamma == None):

    initial_values = array([0.3, 0.1, 0.1])
    boundaries = [(0, 1), (0, 1), (0, 1)]
    type = 'additive'

    parameters = fmin_l_bfgs_b(RMSE, x0 = initial_values, args = (Y, type, m), bounds = boundaries, approx_grad = True)
    alpha, beta, gamma = parameters[0]

  a = [sum(Y[0:m]) / float(m)]
  b = [(sum(Y[m:2 * m]) - sum(Y[0:m])) / m ** 2]
  s = [Y[i] - a[0] for i in range(m)]
  y = [a[0] + b[0] + s[0]]
  rmse = 0

  for i in range(len(Y) + fc):

    if i == len(Y):
      Y.append(a[-1] + b[-1] + s[-m])

    a.append(alpha * (Y[i] - s[i]) + (1 - alpha) * (a[i] + b[i]))
    b.append(beta * (a[i + 1] - a[i]) + (1 - beta) * b[i])
    s.append(gamma * (Y[i] - a[i] - b[i]) + (1 - gamma) * s[i])
    y.append(a[i + 1] + b[i + 1] + s[i + 1])

  rmse = sqrt(sum([(m - n) ** 2 for m, n in zip(Y[:-fc], y[:-fc - 1])]) / len(Y[:-fc]))

  return Y[-fc:], alpha, beta, gamma, rmse

def multiplicative(x, m, fc, alpha = None, beta = None, gamma = None):

  Y = x[:]

  if (alpha == None or beta == None or gamma == None):

    initial_values = array([0.0, 1.0, 0.0])
    boundaries = [(0, 1), (0, 1), (0, 1)]
    type = 'multiplicative'

    parameters = fmin_l_bfgs_b(RMSE, x0 = initial_values, args = (Y, type, m), bounds = boundaries, approx_grad = True)
    alpha, beta, gamma = parameters[0]

  a = [sum(Y[0:m]) / float(m)]
  b = [(sum(Y[m:2 * m]) - sum(Y[0:m])) / m ** 2]
  s = [Y[i] / a[0] for i in range(m)]
  y = [(a[0] + b[0]) * s[0]]
  rmse = 0

  for i in range(len(Y) + fc):

    if i == len(Y):
      Y.append((a[-1] + b[-1]) * s[-m])

    a.append(alpha * (Y[i] / s[i]) + (1 - alpha) * (a[i] + b[i]))
    b.append(beta * (a[i + 1] - a[i]) + (1 - beta) * b[i])
    s.append(gamma * (Y[i] / (a[i] + b[i])) + (1 - gamma) * s[i])
    y.append((a[i + 1] + b[i + 1]) * s[i + 1])

  rmse = sqrt(sum([(m - n) ** 2 for m, n in zip(Y[:-fc], y[:-fc - 1])]) / len(Y[:-fc]))

  return Y[-fc:], alpha, beta, gamma, rmse


from io import open
from setuptools import setup

import sys
sys.lore_no_env = True

import lore


def readme():
    with open('README.rst', 'r', encoding='utf-8') as f:
        return f.read()

sql = ['sqlalchemy>=1.2.0b3, <1.2.99', 'sqlalchemy-migrate>=0.11, <0.11.99']
postgres = ['psycopg2>=2.7, <2.7.99'] + sql
redshift = ['sqlalchemy-redshift>=0.7, <0.7.99'] + sql
redis = ['redis>=2.10, <2.10.99']
s3 = ['boto3>=1.4, <1.4.99']

keras = ['Keras==2.0.9', 'tensorflow>=1.3, <1.4.99', 'dill>=0.2, <0.2.99', 'h5py>=2.7, <2.7.99']
xgboost = ['xgboost>=0.6a2, <0.6.99']
scikit = ['scikit-learn>=0.19, <0.19.99']

all = list(set(keras + xgboost + scikit + postgres + redshift + redis + s3))
devel = all + ['moto>=1.1, <1.1.99']

setup(
    name='lore',
    version=lore.__version__,
    description='a framework for building and using data science',
    long_description=readme(),
    classifiers=[
        lore.__status__,
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Environment :: Console',
    ],
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, <4',
    keywords='machine learning framework tensorflow airflow',
    url='http://github.com/instacart/lore',
    author=lore.__author__,
    author_email=lore.__email__,
    license='MIT',
    packages=[
        'lore',
        'lore.estimators',
        'lore.estimators.holt_winters',
        'lore.features',
        'lore.io',
        'lore.models',
        'lore.stores',
    ],
    install_requires=[
        'future>=0.15, <0.16.99',
        'inflection>=0.3, <0.3.99',
        'jupyter>=1.0, <1.0.99',
        'numpy>=1.14, <1.14.99',
        'pandas>=0.20, <0.22.99',
        'python-dotenv>=0.6, <0.7.99',
        'six>=1.10, <1.11.99',
        'smart-open>=1.5, <1.5.99',
        'tabulate>=0.7.5, <0.8.99',
    ],
    extras_require={
        'all': all,
        'devel': devel,
        'keras': keras,
        'postgres': postgres,
        'redis': redis,
        'redshift': redshift,
        's3': s3,
        'scikit': scikit,
        'xgboost': xgboost,
    },
    zip_safe=True,
    test_suite='tests',
    tests_require=[],
    package_data={
        '': [
            'data/names.csv',
            'template/.*',
            'template/**/*',
            'template/**/.*',
            'template/.**/*',
            'template/.**/.*'
        ]
    },
    entry_points={
        'console_scripts': [
            'lore=lore.__main__:main',
        ],
    },
)



# -*- coding: utf-8 -*-
import datetime
from abc import ABCMeta, abstractmethod, abstractproperty
import pandas as pd
import inflection

class Base(object):
    __metaclass__ = ABCMeta

    def __init__(self):
        self._data = pd.DataFrame()

    @abstractmethod
    def key(self):
        """
        :return: Composite or a single key for index
        """
        pass

    @abstractmethod
    def get_data(self):
        pass

    @abstractmethod
    def publish(self):
        """
        Publish the feature to store ( S3, Memcache, Redis, Cassandra etc)
        :return: None
        """
        pass

    @property
    def version(self, version = str(datetime.date.today())):
        """
        Feature version : Override this method if you want to manage versions yourself
        ex 'v1', 'v2'
        By default will you date as the version information
        :param version:
        :return:
        """
        return version

    def name(self):
        return inflection.underscore(self.__class__.__name)

    def values(self):
        return self._data.columns.values.tolist()

    def _features_as_kv(self):
        """
        Return features row as kv pairs so that they can be stored in memcache or redis and
        used at serving layer
        :return: a nested hash for each column
        """
        self._data = self.get_data()
        key_list = self.key()
        values_list = self.values()
        result = {}
        for column in values_list:
            key_prefix = self.cache_key_prefix() + "#" + column
            self._data['cache_key'] = self._data[key_list].apply(lambda xdf: key_prefix + "=" + '#'.join(xdf.astype(str).values), axis=1)
            result[column] = dict(zip(self._data.cache_key.values, self._data[column].values))
        return result


    def cache_key_prefix(self):
        return ('#').join(self.key())

    def generate_row_keys(self):
        """
        Method for generating key features at serving time or prediction time
        :param data: Pass in the data that is necessary for generating the keys
         Example :
            Feature : User warehouse searches and conversions
            Keys will be of the form 'user_id#warehouse_id#searches=23811676#3'
            Keys will be of the form 'user_id#warehouse_id#conversions=23811676#3'
            data Frame should have values for all the columns as feature_key in this case ['user_id','warehouse_id']
        :return:
        """
        keys = self.key
        columns = self.values
        if not self._data:
            self._data = self.get_data()

        for column in columns:
            key_prefix = self.cache_key_prefix() + "#" + column
            self._data['cache_key'] = self._data[keys].apply(lambda xdf: key_prefix + "=" + '#'.join(xdf.astype(str).values),
                                                 axis=1)
        return list(self._data['cache_key'].values)


    def __repr__(self):
        return (
            """
                Version      : {}
                Name         : {}
                Keys         : {}
                Values       : {}
                Rows         : {}
            """.format(self.version, self.name(), self.key(), self.values(), len(self._data))
            # """.format(self.version, self.name, self.key, self.values, len([1]))
        )

    def metadata(self):
        return {
            "version": self.version,
            "name": self.name(),
            "keys": self.key(),
            "values": self.values(),
            "num_rows": len(self._data)
        }


    def distribute(self, cache):
        """
        Sync features to a key value compliant cache. Should adhere to cache protocol
        :param cache:
        :return: None
        """
        data = self._features_as_kv()
        for key in data.keys():
            cache.batch_set(data[key])




import os
import tempfile
import re
import configparser
import sys
import logging
import redis

import lore
from lore.util import timer
from lore.io.connection import Connection

if not (sys.version_info.major == 3 and sys.version_info.minor >= 6):
    ModuleNotFoundError = ImportError
try:
    import boto3
    from botocore.exceptions import ClientError
except ModuleNotFoundError as e:
    boto3 = False
    ClientError = Exception

logger = logging.getLogger(__name__)


config = lore.env.database_config
if config:
    try:
        for database, url in config.items('DATABASES'):
            vars()[database] = Connection(url=url, name=database)
    except configparser.NoSectionError:
        pass

    for section in config.sections():
        if section == 'DATABASES':
            continue
            
        options = config._sections[section]
        vars()[section.lower()] = Connection(name=section.lower(), **options)

redis_config = lore.env.redis_config

if redis_config:
    try:
        for section in config.sections():
            vars()[section.lower()] = redis.StrictRedis(host=redis_config.get(section, 'url'),
                                                        port=redis_config.get(section, 'port'))
    except:
        pass
else:
    redis_conn = redis.StrictRedis(host='localhost', port=6379)

if boto3:
    config = lore.env.aws_config
    s3 = None
    if config and 'ACCESS_KEY' in config.sections():
        s3 = boto3.resource(
            's3',
            aws_access_key_id=config.get('ACCESS_KEY', 'id'),
            aws_secret_access_key=config.get('ACCESS_KEY', 'secret')
        )
    else:
        s3 = boto3.resource('s3')

    if s3 and config and 'BUCKET' in config.sections():
        bucket = s3.Bucket(config.get('BUCKET', 'name'))


def download(local_path, remote_path=None, cache=True):
    if remote_path is None:
        remote_path = remote_from_local(local_path)
    remote_path = prefix_remote_root(remote_path)

    if cache and os.path.exists(local_path):
        return

    dir = os.path.dirname(local_path)
    if not os.path.exists(dir):
        os.makedirs(dir)
    
    with timer('DOWNLOAD: %s -> %s' % (remote_path, local_path)):
        temp_file, temp_path = tempfile.mkstemp()
        try:
            bucket.download_file(remote_path, temp_path)
        except ClientError as e:
            logger.error("Error downloading file: %s" % e)
            raise
        
        os.rename(temp_path, local_path)


def upload(local_path, remote_path=None):
    if remote_path is None:
        remote_path = remote_from_local(local_path)
    remote_path = prefix_remote_root(remote_path)
    
    with timer('UPLOAD: %s -> %s' % (local_path, remote_path)):
        try:
            bucket.upload_file(local_path, remote_path, ExtraArgs={'ServerSideEncryption': 'AES256'})
        except ClientError as e:
            logger.error("Error uploading file: %s" % e)
            raise


def remote_from_local(local_path):
    return re.sub(
        r'^%s' % re.escape(lore.env.work_dir),
        '',
        local_path
    )


def prefix_remote_root(path):
    if path.startswith('/'):
        path = path[1:]

    if not path.startswith(lore.env.name + '/'):
        path = os.path.join(lore.env.name, path)

    return path
